import torchquantum as tq
import itertools

from typing import List, Iterable

__all__ = [
    "SuperQuantumModule",
    "Super1QLayer",
    "Super2QLayer",
    "Super1QShareFrontLayer",
    "Super1QSingleWireLayer",
    "Super1QAllButOneLayer",
    "Super2QAllShareFrontLayer",
    "Super2QAllLayer",
    "Super2QAlterLayer",
    "super_layer_name_dict",
]


def get_combs(inset: List, n=None) -> List[List]:
    all_combs = []
    if n is None:
        # all possible combinations, with different #elements in a set
        for k in range(1, len(inset) + 1):
            all_combs.extend(list(map(list, itertools.combinations(inset, k))))
    elif isinstance(n, int):
        all_combs.extend(list(map(list, itertools.combinations(inset, n))))
    elif isinstance(n, Iterable):
        for k in n:
            all_combs.extend(list(map(list, itertools.combinations(inset, k))))

    return all_combs


class SuperQuantumModule(tq.QuantumModule):
    def __init__(self, n_wires):
        super().__init__()
        self.n_wires = n_wires
        self.sample_arch = None

    def set_sample_arch(self, sample_arch):
        self.sample_arch = sample_arch

    @property
    def arch_space(self):
        return None

    def count_sample_params(self):
        raise NotImplementedError


class Super1QLayer(SuperQuantumModule):
    def __init__(self, op, n_wires: int, has_params=False, trainable=False):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.sample_wires = None
        self.ops_all = tq.QuantumModuleList()
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_wires):
            if k in self.sample_arch:
                self.ops_all[k](q_device, wires=k)

    @property
    def arch_space(self):
        choices = list(range(self.n_wires))
        return get_combs(choices)

    def count_sample_params(self):
        return len(self.sample_arch) * self.op.num_params


class Super2QLayer(SuperQuantumModule):
    def __init__(
        self, op, n_wires: int, has_params=False, trainable=False, wire_reverse=False
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.ops_all = tq.QuantumModuleList()

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_wires):
            if [k, (k + 1) % self.n_wires] in self.sample_arch or [
                (k + 1) % self.n_wires,
                k,
            ] in self.sample_arch:
                wires = sorted([k, (k + 1) % self.n_wires], reverse=self.wire_reverse)
                self.ops_all[k](q_device, wires=wires)

    @property
    def arch_space(self):
        choices = list(map(list, itertools.combinations(list(range(self.n_wires)), 2)))
        return get_combs(choices)

    def count_sample_params(self):
        return len(self.sample_arch) * self.op.num_params


class Super1QShareFrontLayer(SuperQuantumModule):
    """Share the front wires, the rest can be added"""

    def __init__(
        self,
        op,
        n_wires: int,
        n_front_share_wires: int,
        has_params=False,
        trainable=False,
    ):
        super().__init__(n_wires=n_wires)
        self.n_front_share_wires = n_front_share_wires
        self.op = op
        self.ops_all = tq.QuantumModuleList()
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        self.q_device = q_device
        for k in range(self.n_wires):
            if k < self.sample_arch:
                self.ops_all[k](q_device, wires=k)

    @property
    def arch_space(self):
        return list(range(self.n_front_share_wires, self.n_wires + 1))

    def count_sample_params(self):
        return min(self.sample_arch, len(self.ops_all)) * self.op.num_params


class Super1QSingleWireLayer(SuperQuantumModule):
    """Only one wire will have a gate"""

    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.ops_all = tq.QuantumModuleList()
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_wires):
            if k == self.sample_arch:
                self.ops_all[k](q_device, wires=k)

    @property
    def arch_space(self):
        return list(range(self.n_wires))

    def count_sample_params(self):
        return self.op.num_params


class Super1QAllButOneLayer(SuperQuantumModule):
    """Only one wire will NOT have the gate"""

    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.ops_all = tq.QuantumModuleList()
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_wires):
            if k != self.sample_arch:
                self.ops_all[k](q_device, wires=k)

    @property
    def arch_space(self):
        return list(range(self.n_wires))

    def count_sample_params(self):
        return (self.n_wires - 1) * self.op.num_params


class Super2QAllShareFrontLayer(SuperQuantumModule):
    """pattern:
    circular = False
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5]
    jump = 3: [0, 3], [1, 4], [2, 5]
    jump = 4: [0, 4], [1, 5]
    jump = 5: [0, 5]

    circular = True
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5], [4, 0], [5, 1]
    jump = 3: [0, 3], [1, 4], [2, 5], [3, 0], [4, 1], [5, 2]
    jump = 4: [0, 4], [1, 5], [2, 0], [3, 1], [4, 2], [5, 3]
    jump = 5: [0, 5], [1, 0], [2, 1], [3, 2], [4, 3], [5, 4]
    """

    def __init__(
        self,
        op,
        n_wires: int,
        n_front_share_ops: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
        circular=False,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.n_front_share_ops = n_front_share_ops
        self.jump = jump
        self.circular = circular

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        self.ops_all = tq.QuantumModuleList()
        if circular:
            self.n_ops = n_wires
        else:
            self.n_ops = n_wires - jump
        for k in range(self.n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_ops):
            if k < self.sample_arch:
                wires = [k, (k + self.jump) % self.n_wires]
                if self.wire_reverse:
                    wires.reverse()
                self.ops_all[k](q_device, wires=wires)

    @property
    def arch_space(self):
        choices = list(range(self.n_front_share_ops, self.n_ops + 1))
        return choices

    def count_sample_params(self):
        return min(self.sample_arch, self.n_ops) * self.op.num_params


class Super2QAllLayer(SuperQuantumModule):
    """pattern:
    circular = False
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5]
    jump = 3: [0, 3], [1, 4], [2, 5]
    jump = 4: [0, 4], [1, 5]
    jump = 5: [0, 5]

    circular = True
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5], [4, 0], [5, 1]
    jump = 3: [0, 3], [1, 4], [2, 5], [3, 0], [4, 1], [5, 2]
    jump = 4: [0, 4], [1, 5], [2, 0], [3, 1], [4, 2], [5, 3]
    jump = 5: [0, 5], [1, 0], [2, 1], [3, 2], [4, 3], [5, 4]
    """

    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
        circular=False,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.jump = jump
        self.circular = circular

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        self.ops_all = tq.QuantumModuleList()
        if circular:
            self.n_ops = n_wires
        else:
            self.n_ops = n_wires - jump
        for k in range(self.n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_ops):
            wires = [k, (k + self.jump) % self.n_wires]
            if self.wire_reverse:
                wires.reverse()

            if wires in self.sample_arch or list(reversed(wires)) in self.sample_arch:
                self.ops_all[k](q_device, wires=wires)

    @property
    def arch_space(self):
        choices = []
        for k in range(self.n_ops):
            wires = [k, (k + self.jump) % self.n_wires]
            choices.append(wires)
        return get_combs(choices)

    def count_sample_params(self):
        return len(self.sample_arch) * self.op.num_params


class Super2QAlterLayer(SuperQuantumModule):
    """pattern
    jump = 1: [0, 1], [2, 3], [4, 5], [1, 2], [3, 4], [5, 6]
    jump = 2: [0, 2], [4, 6], [2, 4]
    jump = 3: [0, 3], [3, 6]
    jump = 4: [0, 4]
    jump = 5: [0, 5]
    jump = 6: [0, 6]
    """

    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.jump = jump

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        self.ops_all = tq.QuantumModuleList()
        self.n_ops = (n_wires - 1) // jump
        for k in range(self.n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))
        self.wires_choices = []
        k = 0
        while k < n_wires:
            if k < n_wires and k + jump < n_wires:
                self.wires_choices.append([k, k + jump])
            k += jump * 2

        k = jump
        while k < n_wires:
            if k < n_wires and k + jump < n_wires:
                self.wires_choices.append([k, k + jump])
            k += jump * 2

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_ops):
            wires = self.wires_choices[k]
            if self.wire_reverse:
                wires.reverse()

            if wires in self.sample_arch or list(reversed(wires)) in self.sample_arch:
                self.ops_all[k](q_device, wires=wires)

    @property
    def arch_space(self):
        return get_combs(self.wires_choices)

    def count_sample_params(self):
        return len(self.sample_arch) * self.op.num_params


class Super2QAlterShareFrontLayer(SuperQuantumModule):
    """pattern
    jump = 1: [0, 1], [2, 3], [4, 5], [1, 2], [3, 4], [5, 6]
    jump = 2: [0, 2], [4, 6], [2, 4]
    jump = 3: [0, 3], [3, 6]
    jump = 4: [0, 4]
    jump = 5: [0, 5]
    jump = 6: [0, 6]
    """

    def __init__(
        self,
        op,
        n_wires: int,
        n_front_share_ops: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
    ):
        super().__init__(n_wires=n_wires)
        self.op = op
        self.n_front_share_ops = n_front_share_ops
        self.jump = jump

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        self.ops_all = tq.QuantumModuleList()
        self.n_ops = (n_wires - 1) // jump
        for k in range(self.n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))
        self.wires_choices = []
        k = 0
        while k < n_wires:
            if k < n_wires and k + jump < n_wires:
                self.wires_choices.append([k, k + jump])
            k += jump * 2

        k = jump
        while k < n_wires:
            if k < n_wires and k + jump < n_wires:
                self.wires_choices.append([k, k + jump])
            k += jump * 2

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_ops):
            if k < self.sample_arch:
                wires = self.wires_choices[k]
                if self.wire_reverse:
                    wires.reverse()

                self.ops_all[k](q_device, wires=wires)

    @property
    def arch_space(self):
        choices = list(range(self.n_front_share_ops, self.n_ops + 1))
        return choices

    def count_sample_params(self):
        return min(self.sample_arch, self.n_ops) * self.op.num_params


class SuperLayerTemplate0(SuperQuantumModule):
    def __init__(self, arch: dict = None):
        super().__init__(n_wires=arch["n_wires"])
        self.arch = arch

        self.n_front_share_wires = arch.get("n_front_share_wires", None)
        self.n_front_share_ops = arch.get("n_front_share_ops", None)

        self.n_blocks = arch.get("n_blocks", None)
        self.n_layers_per_block = arch.get("n_layers_per_block", None)
        self.n_front_share_blocks = arch.get("n_front_share_blocks", None)

        self.sample_n_blocks = None

        self.super_layers_all = self.build_super_layers()

    def build_super_layers(self):
        raise NotImplementedError

    def set_sample_arch(self, sample_arch):
        for k, layer_arch in enumerate(sample_arch[:-1]):
            self.super_layers_all[k].set_sample_arch(layer_arch)
        self.sample_n_blocks = sample_arch[-1]

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        self.q_device = q_device
        for k in range(len(self.super_layers_all)):
            if k < self.sample_n_blocks * self.n_layers_per_block:
                self.super_layers_all[k](q_device)

    def count_sample_params(self):
        n_params = 0
        for layer_idx, layer in enumerate(self.super_layers_all):
            if layer_idx < self.sample_n_blocks * self.n_layers_per_block:
                n_params += layer.count_sample_params()
        return n_params


class SuperU3CU3ShareFrontLayer0(SuperLayerTemplate0):
    """u3 cu3 blocks"""

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.U3,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super2QAllShareFrontLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return super_layers_all


class SuperU3CU3ArbitraryLayer0(SuperLayerTemplate0):
    """u3 cu3 blocks arbitrary n gates"""

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QLayer(
                    op=tq.U3, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                Super2QAllLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return super_layers_all


class SuperIBMBasisShareFrontLayer0(SuperLayerTemplate0):
    """cnot rz sx x share front blocks"""

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RZ,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.PauliX,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RZ,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.SX,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RZ,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super2QAllShareFrontLayer(
                    op=tq.CNOT,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    jump=1,
                    circular=True,
                )
            )
        return super_layers_all


class SuperIBMBasisArbitraryLayer0(SuperLayerTemplate0):
    """cnot rz sx x blocks arbitrary n gates"""

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(Super1QLayer(op=tq.PauliX, n_wires=self.n_wires))
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(Super1QLayer(op=tq.SX, n_wires=self.n_wires))
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                Super2QAllLayer(op=tq.CNOT, n_wires=self.n_wires, jump=1, circular=True)
            )
        return super_layers_all


class SuperSethArbitraryLayer0(SuperLayerTemplate0):
    """
    zz and ry blocks arbitrary n gates, from Seth Lloyd paper
    https://arxiv.org/pdf/2001.03622.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super2QAllLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
        return super_layers_all


class SuperSethShareFrontLayer0(SuperLayerTemplate0):
    """
    zz and ry blocks share front n gates, from Seth Lloyd paper
    https://arxiv.org/pdf/2001.03622.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super2QAllShareFrontLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RY,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
        return super_layers_all


class SuperBarrenArbitraryLayer0(SuperLayerTemplate0):
    """
    rx ry rz and cz blocks arbitrary n gates, from Barren plateaus paper
    https://arxiv.org/pdf/1803.11173.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        super_layers_all.append(Super1QLayer(op=tq.SHadamard, n_wires=self.n_wires))

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RX, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                Super1QLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                Super2QAlterLayer(op=tq.CZ, n_wires=self.n_wires, jump=1)
            )
        return super_layers_all


class SuperBarrenShareFrontLayer0(SuperLayerTemplate0):
    """
    rx ry rz and cz blocks share front n gates, from Barren plateaus paper
    https://arxiv.org/pdf/1803.11173.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        super_layers_all.append(
            Super1QShareFrontLayer(
                op=tq.SHadamard,
                n_wires=self.n_wires,
                n_front_share_wires=self.n_front_share_wires,
            )
        )

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RX,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RY,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super1QShareFrontLayer(
                    op=tq.RZ,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                Super2QAlterShareFrontLayer(
                    op=tq.CZ,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    jump=1,
                )
            )
        return super_layers_all


class SuperFarhiArbitraryLayer0(SuperLayerTemplate0):
    """
    zx and xx blocks arbitrary n gates, from Farhi paper
    https://arxiv.org/pdf/1802.06002.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.RZX,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.RXX,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return super_layers_all


class SuperFarhiShareFrontLayer0(SuperLayerTemplate0):
    """
    zx and xx blocks share front n gates, from Farhi paper
    https://arxiv.org/pdf/1802.06002.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.RZX,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.RXX,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return super_layers_all


class SuperMaxwellArbitraryLayer0(SuperLayerTemplate0):
    """
    rx, s, cnot, ry, t, swap, rz, h, sswap, u1, cu3,
    blocks arbitrary n gates, from Maxwell paper
    https://arxiv.org/pdf/1904.04767.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                tq.Super1QLayer(
                    op=tq.RX, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(tq.Super1QLayer(op=tq.S, n_wires=self.n_wires))
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.CNOT, n_wires=self.n_wires, jump=1, circular=True
                )
            )

            super_layers_all.append(
                tq.Super1QLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(tq.Super1QLayer(op=tq.T, n_wires=self.n_wires))
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.SWAP, n_wires=self.n_wires, jump=1, circular=True
                )
            )

            super_layers_all.append(
                tq.Super1QLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(tq.Super1QLayer(op=tq.T, n_wires=self.n_wires))
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.SSWAP, n_wires=self.n_wires, jump=1, circular=True
                )
            )

            super_layers_all.append(
                tq.Super1QLayer(
                    op=tq.U1, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            super_layers_all.append(
                tq.Super2QAllLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )

        return super_layers_all


class SuperMaxwellShareFrontLayer0(SuperLayerTemplate0):
    """
    rx, s, cnot, ry, t, swap, rz, h, sswap, u1, cu3,
    blocks share front n gates, from Maxwell paper
    https://arxiv.org/pdf/1904.04767.pdf
    """

    def build_super_layers(self):
        super_layers_all = tq.QuantumModuleList()

        for k in range(self.arch["n_blocks"]):
            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.RX,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.S,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                )
            )
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.CNOT,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    jump=1,
                    circular=True,
                )
            )

            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.RY,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.T,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                )
            )
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.SWAP,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    jump=1,
                    circular=True,
                )
            )

            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.RZ,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.T,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                )
            )
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.SSWAP,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    jump=1,
                    circular=True,
                )
            )

            super_layers_all.append(
                tq.Super1QShareFrontLayer(
                    op=tq.U1,
                    n_wires=self.n_wires,
                    n_front_share_wires=self.n_front_share_wires,
                    has_params=True,
                    trainable=True,
                )
            )
            super_layers_all.append(
                tq.Super2QAllShareFrontLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    n_front_share_ops=self.n_front_share_ops,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )

        return super_layers_all


super_layer_name_dict = {
    "u3cu3_s0": SuperU3CU3ShareFrontLayer0,
    "seth_s0": SuperSethShareFrontLayer0,
    "barren_s0": SuperBarrenShareFrontLayer0,
    "farhi_s0": SuperFarhiShareFrontLayer0,
    "maxwell_s0": SuperMaxwellShareFrontLayer0,
    "ibmbasis_s0": SuperIBMBasisShareFrontLayer0,
    "u3cu3_a0": SuperU3CU3ArbitraryLayer0,
    "seth_a0": SuperSethArbitraryLayer0,
    "barren_a0": SuperBarrenArbitraryLayer0,
    "farhi_a0": SuperFarhiArbitraryLayer0,
    "maxwell_a0": SuperMaxwellArbitraryLayer0,
    "ibmbasis_a0": SuperIBMBasisArbitraryLayer0,
}



import torchquantum as tq
import torchquantum.functional as tqf
import torch
import torch.nn.functional as F
import numpy as np
import os
import random
import datetime

from torchquantum.encoding import encoder_op_list_name_dict
from torchpack.utils.logging import logger
from torchquantum.layers import layer_name_dict
from torchpack.utils.config import configs
import pdb


class QVQEModel0(tq.QuantumModule):
    def __init__(self, arch, hamil_info):
        super().__init__()
        self.arch = arch
        self.hamil_info = hamil_info
        self.n_wires = arch["n_wires"]
        self.q_device = tq.QuantumDevice(n_wires=self.n_wires)
        self.q_layer = layer_name_dict[arch["q_layer_name"]](arch)
        self.measure = tq.MeasureMultipleTimes(obs_list=hamil_info["hamil_list"])
        self.num_forwards = 0
        self.n_params = len(list(self.parameters()))

    def forward(self, x, verbose=False, use_qiskit=False):
        if use_qiskit:
            x = self.qiskit_processor.process_multi_measure(
                self.q_device, self.q_layer, self.measure
            )
        else:
            self.q_device.reset_states(bsz=1)
            self.q_layer(self.q_device)
            x = self.measure(self.q_device)

        hamil_coefficients = torch.tensor(
            [hamil["coefficient"] for hamil in self.hamil_info["hamil_list"]],
            device=x.device,
        ).double()

        for k, hamil in enumerate(self.hamil_info["hamil_list"]):
            for wire, observable in zip(hamil["wires"], hamil["observables"]):
                if observable == "i":
                    x[k][wire] = 1
            for wire in range(self.q_device.n_wires):
                if wire not in hamil["wires"]:
                    x[k][wire] = 1

        if verbose:
            logger.info(f"[use_qiskit]={use_qiskit}, expectation:\n {x.data}")

        x = torch.cumprod(x, dim=-1)[:, -1].double()
        x = torch.dot(x, hamil_coefficients)

        if x.dim() == 0:
            x = x.unsqueeze(0)

        return x

    def shift_and_run(
        self, x, global_step, total_step, verbose=False, use_qiskit=False
    ):
        with torch.no_grad():
            if use_qiskit:
                self.q_device.reset_states(bsz=1)
                x = self.qiskit_processor.process_multi_measure(
                    self.q_device, self.q_layer, self.measure
                )
                self.grad_list = []
                for i, param in enumerate(self.parameters()):
                    param.copy_(param + np.pi * 0.5)
                    out1 = self.qiskit_processor.process_multi_measure(
                        self.q_device, self.q_layer, self.measure
                    )
                    param.copy_(param - np.pi)
                    out2 = self.qiskit_processor.process_multi_measure(
                        self.q_device, self.q_layer, self.measure
                    )
                    param.copy_(param + np.pi * 0.5)
                    grad = 0.5 * (out1 - out2)
                    self.grad_list.append(grad)
            else:
                self.q_device.reset_states(bsz=1)
                self.q_layer(self.q_device)
                x = self.measure(self.q_device)
                self.grad_list = []
                for i, param in enumerate(self.parameters()):
                    param.copy_(param + np.pi * 0.5)
                    self.q_device.reset_states(bsz=1)
                    self.q_layer(self.q_device)
                    out1 = self.measure(self.q_device)
                    param.copy_(param - np.pi)
                    self.q_device.reset_states(bsz=1)
                    self.q_layer(self.q_device)
                    out2 = self.measure(self.q_device)
                    param.copy_(param + np.pi * 0.5)
                    grad = 0.5 * (out1 - out2)
                    self.grad_list.append(grad)

        hamil_coefficients = torch.tensor(
            [hamil["coefficient"] for hamil in self.hamil_info["hamil_list"]],
            device=x.device,
        ).double()
        x_axis = []
        y_axis = []
        for k, hamil in enumerate(self.hamil_info["hamil_list"]):
            for wire, observable in zip(hamil["wires"], hamil["observables"]):
                if observable == "i":
                    x[k][wire] = 1
                    x_axis.append(k)
                    y_axis.append(wire)
            for wire in range(self.q_device.n_wires):
                if wire not in hamil["wires"]:
                    x[k][wire] = 1
                    x_axis.append(k)
                    y_axis.append(wire)
        for grad in self.grad_list:
            grad[x_axis, y_axis] = 0

        self.circuit_output = x
        self.circuit_output.requires_grad = True
        self.num_forwards += (1 + 2 * self.n_params) * x.shape[0]
        if verbose:
            logger.info(f"[use_qiskit]={use_qiskit}, expectation:\n {x.data}")

        x = torch.cumprod(x, dim=-1)[:, -1].double()
        x = torch.dot(x, hamil_coefficients)

        if x.dim() == 0:
            x = x.unsqueeze(0)

        return x

    def backprop_grad(self):
        for i, param in enumerate(self.parameters()):
            param.grad = (
                torch.sum(self.grad_list[i] * self.circuit_output.grad)
                .to(dtype=torch.float32)
                .view(param.shape)
            )


model_dict = {
    "vqe_0": QVQEModel0,
}



"""
Description:
Author: Jiaqi Gu (jqgu@utexas.edu)
Date: 2021-04-04 13:38:43
LastEditors: Jiaqi Gu (jqgu@utexas.edu)
LastEditTime: 2021-04-04 15:16:55
"""

import os
import numpy as np
import torch

from torch import Tensor
from torchpack.datasets.dataset import Dataset
from torchvision import transforms
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import download_url
from typing import Any, Callable, Dict, List, Optional, Tuple


__all__ = ["Vowel"]


class VowelRecognition(VisionDataset):
    url = (
        "http://archive.ics.uci.edu/ml/machine-learning-databases"
        "/undocumented/connectionist-bench/vowel/vowel-context.data"
    )
    filename = "vowel-context.data"
    folder = "vowel"

    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        n_features: int = 10,
        train_ratio: float = 0.7,
        download: bool = False,
    ) -> None:
        root = os.path.join(os.path.expanduser(root), self.folder)
        if transform is None:
            transform = transforms.Compose([transforms.ToTensor()])
        super(VowelRecognition, self).__init__(
            root, transform=transform, target_transform=target_transform
        )

        self.train = train  # training set or test set
        self.train_ratio = train_ratio

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        self.n_features = n_features
        assert 1 <= n_features <= 10, print(
            f"Only support maximum 10 features, but got{n_features}"
        )
        self.data: Any = []
        self.targets = []

        self.process_raw_data()
        self.data, self.targets = self.load(train=train)

    def process_raw_data(self) -> None:
        processed_dir = os.path.join(self.root, "processed")
        processed_training_file = os.path.join(processed_dir, "training.pt")
        processed_test_file = os.path.join(processed_dir, "test.pt")
        if os.path.exists(processed_training_file) and os.path.exists(
            processed_test_file
        ):
            with open(os.path.join(self.root, "processed/training.pt"), "rb") as f:
                data, targets = torch.load(f)
                if data.shape[-1] == self.n_features:
                    print("Data already processed")
                    return
        data, targets = self._load_dataset()
        data_train, targets_train, data_test, targets_test = self._split_dataset(
            data, targets
        )
        data_train, data_test = self._preprocess_dataset(data_train, data_test)
        self._save_dataset(
            data_train, targets_train, data_test, targets_test, processed_dir
        )

    def _load_dataset(self) -> Tuple[torch.Tensor, torch.Tensor]:
        data = []
        targets = []
        with open(os.path.join(self.root, "raw", self.filename), "r") as f:
            for line in f:
                line = line.strip().split()[3:]
                label = int(line[-1])
                targets.append(label)
                example = [float(i) for i in line[:-1]]
                data.append(example)

            data = torch.Tensor(data)
            targets = torch.LongTensor(targets)
        return data, targets

    def _split_dataset(self, data: Tensor, targets: Tensor) -> Tuple[Tensor, ...]:
        from sklearn.model_selection import train_test_split

        data_train, data_test, targets_train, targets_test = train_test_split(
            data, targets, train_size=self.train_ratio, random_state=42
        )
        print(
            f"training: {data_train.shape[0]} examples, "
            f"test: {data_test.shape[0]} examples"
        )
        return data_train, targets_train, data_test, targets_test

    def _preprocess_dataset(
        self, data_train: Tensor, data_test: Tensor
    ) -> Tuple[Tensor, Tensor]:
        from sklearn.decomposition import PCA
        from sklearn.preprocessing import MinMaxScaler, RobustScaler

        pca = PCA(n_components=self.n_features)
        data_train_reduced = pca.fit_transform(data_train)
        data_test_reduced = pca.transform(data_test)

        rs = RobustScaler(quantile_range=(10, 90)).fit(
            np.concatenate([data_train_reduced, data_test_reduced], 0)
        )
        data_train_reduced = rs.transform(data_train_reduced)
        data_test_reduced = rs.transform(data_test_reduced)
        mms = MinMaxScaler()
        mms.fit(np.concatenate([data_train_reduced, data_test_reduced], 0))
        data_train_reduced = mms.transform(data_train_reduced)
        data_test_reduced = mms.transform(data_test_reduced)

        return (
            torch.from_numpy(data_train_reduced).float(),
            torch.from_numpy(data_test_reduced).float(),
        )

    @staticmethod
    def _save_dataset(
        data_train: Tensor,
        targets_train: Tensor,
        data_test: Tensor,
        targets_test: Tensor,
        processed_dir: str,
    ) -> None:
        os.makedirs(processed_dir, exist_ok=True)
        processed_training_file = os.path.join(processed_dir, "training.pt")
        processed_test_file = os.path.join(processed_dir, "test.pt")
        with open(processed_training_file, "wb") as f:
            torch.save((data_train, targets_train), f)

        with open(processed_test_file, "wb") as f:
            torch.save((data_test, targets_test), f)
        print(f"Processed dataset saved")

    def load(self, train: bool = True):
        filename = "training.pt" if train else "test.pt"
        with open(os.path.join(self.root, "processed", filename), "rb") as f:
            data, targets = torch.load(f)
            if isinstance(data, np.ndarray):
                data = torch.from_numpy(data)
            if isinstance(targets, np.ndarray):
                targets = torch.from_numpy(targets)
        return data, targets

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return
        download_url(
            self.url, root=os.path.join(self.root, "raw"), filename=self.filename
        )

    def _check_integrity(self) -> bool:
        return os.path.exists(os.path.join(self.root, "raw", self.filename))

    def __len__(self):
        return self.targets.size(0)

    def __getitem__(self, item):
        return self.data[item], self.targets[item]

    def extra_repr(self) -> str:
        return "Split: {}".format("Train" if self.train is True else "Test")


class VowelRecognitionDataset:
    def __init__(
        self,
        root: str,
        split: str,
        test_ratio: float,
        train_valid_split_ratio: List[float],
        resize: int,
        binarize: bool,
        binarize_threshold: float,
        digits_of_interest: List[int],
    ):
        self.root = root
        self.split = split
        self.test_ratio = test_ratio
        assert 0 < test_ratio < 1, print(
            f"Only support test_ratio from (0, 1), but got {test_ratio}"
        )
        self.train_valid_split_ratio = train_valid_split_ratio
        self.data = None
        self.resize = resize
        self.binarize = binarize
        self.binarize_threshold = binarize_threshold
        self.digits_of_interest = digits_of_interest

        self.load()
        self.n_instance = len(self.data)

    def load(self):
        tran = [transforms.ToTensor()]
        transform = transforms.Compose(tran)

        if self.split == "train" or self.split == "valid":
            train_valid = VowelRecognition(
                self.root,
                train=True,
                download=True,
                transform=transform,
                n_features=self.resize,
                train_ratio=1 - self.test_ratio,
            )
            idx, _ = torch.stack(
                [train_valid.targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            train_valid.targets = train_valid.targets[idx]
            train_valid.data = train_valid.data[idx]

            train_len = int(self.train_valid_split_ratio[0] * len(train_valid))
            split = [train_len, len(train_valid) - train_len]
            train_subset, valid_subset = torch.utils.data.random_split(
                train_valid, split, generator=torch.Generator().manual_seed(1)
            )

            if self.split == "train":
                self.data = train_subset
            else:
                self.data = valid_subset

        else:
            test = VowelRecognition(
                self.root,
                train=False,
                download=True,
                transform=transform,
                n_features=self.resize,
                train_ratio=1 - self.test_ratio,
            )
            idx, _ = torch.stack(
                [test.targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            test.targets = test.targets[idx]
            test.data = test.data[idx]

            self.data = test

    def __getitem__(self, index: int) -> Dict[str, Tensor]:
        data = self.data[index][0]
        if self.binarize:
            data = 1.0 * (data > self.binarize_threshold) + -1.0 * (
                data <= self.binarize_threshold
            )

        digit = self.digits_of_interest.index(self.data[index][1])
        instance = {"vowel": data, "digit": digit}
        return instance

    def __len__(self) -> int:
        return len(self.data)

    def __call__(self, index: int) -> Dict[str, Tensor]:
        return self.__getitem__(index)


class Vowel(Dataset):
    def __init__(
        self,
        root: str,
        test_ratio: float,
        train_valid_split_ratio: List[float],
        resize=28,
        binarize=False,
        binarize_threshold=0.1307,
        digits_of_interest=tuple(range(10)),
    ):
        self.root = root

        super().__init__(
            {
                split: VowelRecognitionDataset(
                    root=root,
                    split=split,
                    test_ratio=test_ratio,
                    train_valid_split_ratio=train_valid_split_ratio,
                    resize=resize,
                    binarize=binarize,
                    binarize_threshold=binarize_threshold,
                    digits_of_interest=digits_of_interest,
                )
                for split in ["train", "valid", "test"]
            }
        )


def test_vowel():
    import pdb

    pdb.set_trace()
    vowel = VowelRecognition(root=".", download=True, n_features=6)
    print(vowel.data.size(), vowel.targets.size())
    vowel = VowelRecognitionDataset(
        root=".",
        split="train",
        test_ratio=0.3,
        train_valid_split_ratio=[0.9, 0.1],
        resize=8,
        binarize=0,
        binarize_threshold=0,
        digits_of_interest=tuple(range(10)),
    )
    # digits_of_interest=(3, 6))
    print(vowel(20))


if __name__ == "__main__":
    test_vowel()



import argparse
import pdb
import torch
import torchquantum as tq
import numpy as np

from torchpack.utils.logging import logger
from torchquantum.operators import op_name_dict
from torchquantum.functional import func_name_dict
from torchquantum.macro import F_DTYPE
from torchquantum.plugins.qiskit_macros import (
    QISKIT_INCOMPATIBLE_FUNC_NAMES,
    QISKIT_INCOMPATIBLE_OPS,
)


class QLayer(tq.QuantumModule):
    def __init__(
        self,
        n_wires,
        wires,
        n_ops_rd,
        n_ops_cin,
        n_funcs,
        qiskit_compatible=False,
    ):
        super().__init__()
        self.n_wires = n_wires
        self.wires = wires
        self.n_ops_rd = n_ops_rd
        self.n_ops_cin = n_ops_cin
        self.n_funcs = n_funcs
        self.rd_layer = None
        if self.n_ops_rd > 0:
            self.rd_layer = tq.RandomLayerAllTypes(
                n_ops=n_ops_rd, wires=wires, qiskit_compatible=qiskit_compatible
            )

        self.cin_op_types = [
            tq.RX,
            tq.RY,
            tq.RZ,
            tq.PhaseShift,
            tq.Rot,
            tq.MultiRZ,
            tq.CRX,
            tq.CRY,
            tq.CRZ,
            tq.CRot,
            tq.U1,
            tq.U2,
            tq.U3,
        ]
        self.funcs = [
            "hadamard",
            "paulix",
            "pauliy",
            "pauliz",
            "s",
            "t",
            "sx",
            "cnot",
            "cz",
            "cy",
            "rx",
            "ry",
            "rz",
            "swap",
            "cswap",
            "toffoli",
            "phaseshift",
            "rot",
            "multirz",
            "crx",
            "cry",
            "crz",
            "crot",
            "u1",
            "u2",
            "u3",
            "qubitunitary",
            "qubitunitaryfast",
            "multicnot",
            "multixcnot",
        ]

        if qiskit_compatible:
            for op in QISKIT_INCOMPATIBLE_OPS:
                self.cin_op_types.remove(op)
            for func in QISKIT_INCOMPATIBLE_FUNC_NAMES:
                self.funcs.remove(func)

        self.x_idx = 0
        self.cin_op_list = tq.QuantumModuleList()
        self.rand_mat = np.random.randn(2**self.n_wires, 2**self.n_wires)
        self.build_random_cin_layer()

        self.func_list = []
        self.func_wires_list = []
        self.build_random_funcs()

        self.cin_op_inverse = np.random.randint(2, size=n_ops_cin)
        self.func_inverse = np.random.randint(2, size=n_funcs)

    def build_random_cin_layer(self):
        cnt = 0
        while cnt < self.n_ops_cin:
            op = np.random.choice(self.cin_op_types)
            n_op_wires = op.num_wires
            if n_op_wires > self.n_wires:
                continue
            if n_op_wires == -1:
                is_AnyWire = True
                n_op_wires = self.n_wires
            else:
                is_AnyWire = False

            op_wires = list(
                np.random.choice(self.wires, size=n_op_wires, replace=False)
            )

            if is_AnyWire:
                operation = op(n_wires=n_op_wires, wires=op_wires)
            else:
                operation = op(wires=op_wires)

            self.cin_op_list.append(operation)
            cnt += 1

    def build_random_funcs(self):
        cnt = 0
        while cnt < self.n_funcs:
            func = np.random.choice(self.funcs)
            n_func_wires = op_name_dict[func]().num_wires
            if n_func_wires > self.n_wires:
                continue
            cnt += 1

            if n_func_wires == -1:
                n_func_wires = self.n_wires

            func_wires = list(
                np.random.choice(self.wires, size=n_func_wires, replace=False)
            )
            self.func_list.append(func)
            self.func_wires_list.append(func_wires)

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice, x):
        self.q_device = q_device

        # test random generated parameterized layers
        if self.n_ops_rd > 0:
            self.rd_layer(q_device)

        # test layer with classical inputs
        for op, is_inverse in zip(self.cin_op_list, self.cin_op_inverse):
            n_op_params = op.num_params
            if n_op_params == -1:
                n_op_params = 2
            params = x[:, self.x_idx : self.x_idx + n_op_params]
            self.x_idx += n_op_params
            op(self.q_device, params=params, inverse=is_inverse)

        # test functional layers
        for func, func_wires, is_inverse in zip(
            self.func_list, self.func_wires_list, self.func_inverse
        ):
            n_func_wires = len(func_wires)
            n_func_params = op_name_dict[func]().num_params

            if n_func_params == 0:
                if func in ["multicnot", "multixcnot"]:
                    func_name_dict[func](
                        self.q_device,
                        wires=func_wires,
                        n_wires=n_func_wires,
                        static=self.static_mode,
                        parent_graph=self.graph,
                        inverse=is_inverse,
                    )
                else:
                    func_name_dict[func](
                        self.q_device,
                        wires=func_wires,
                        static=self.static_mode,
                        parent_graph=self.graph,
                        inverse=is_inverse,
                    )
            elif n_func_params == -1:
                # qubitunitary:
                if func in ["qubitunitary", "qubitunitaryfast", "qubitunitarystrict"]:
                    u, s, v = np.linalg.svd(
                        self.rand_mat[: 2**n_func_wires, : 2**n_func_wires]
                    )
                    params = u @ v
                    func_name_dict[func](
                        self.q_device,
                        wires=func_wires,
                        n_wires=n_func_wires,
                        params=params,
                        static=self.static_mode,
                        parent_graph=self.graph,
                        inverse=is_inverse,
                    )
                else:
                    raise NotImplementedError
            else:
                params = x[:, self.x_idx : self.x_idx + n_func_params]
                self.x_idx += n_func_params
                if func in ["multirz"]:
                    func_name_dict[func](
                        self.q_device,
                        wires=func_wires,
                        n_wires=n_func_wires,
                        params=params,
                        static=self.static_mode,
                        parent_graph=self.graph,
                        inverse=is_inverse,
                    )
                else:
                    func_name_dict[func](
                        self.q_device,
                        wires=func_wires,
                        params=params,
                        static=self.static_mode,
                        parent_graph=self.graph,
                        inverse=is_inverse,
                    )

        self.x_idx = 0


def static_mode_dynamic_vs_static_gradients_test():
    bsz = 7

    for n_wires in range(2, 10):
        x = torch.randn((1, 100000), dtype=F_DTYPE)
        q_dev0 = tq.QuantumDevice(n_wires=n_wires)
        q_dev1 = tq.QuantumDevice(n_wires=n_wires)
        q_dev0.reset_all_eq_states(bsz)
        q_layer = QLayer(
            n_wires=n_wires,
            wires=list(range(n_wires)),
            n_ops_rd=500,
            n_ops_cin=500,
            n_funcs=500,
        )
        q_layer(q_dev0, x)
        loss = q_dev0.states.abs().sum()
        loss.backward()
        q_layer_dynamic_grads = {}
        for name, params in q_layer.named_parameters():
            q_layer_dynamic_grads[name] = params.grad.data.clone()
            params.grad = None

        for wires_per_block in range(1, n_wires + 1):
            q_dev1.reset_all_eq_states(bsz)
            q_layer.static_off()
            q_layer.static_on(wires_per_block=wires_per_block)
            q_layer(q_dev1, x)
            loss = q_dev1.states.abs().sum()
            loss.backward()
            q_layer_static_grads = {}
            for name, params in q_layer.named_parameters():
                q_layer_static_grads[name] = params.grad.data.clone()
                params.grad = None

            name = None
            try:
                for name, _ in q_layer.named_parameters():
                    diff = q_layer_dynamic_grads[name] - q_layer_static_grads[name]
                    assert torch.allclose(
                        q_layer_dynamic_grads[name],
                        q_layer_static_grads[name],
                        atol=1e-4,
                    )
                    logger.info(f"Diff: {diff}")
                    logger.info(
                        f"PASS: [n_wires]={n_wires}, dynamic VS "
                        f"static run with [wires_per_block]="
                        f"{wires_per_block}. Params: {name}"
                    )
            except AssertionError:
                logger.exception(
                    f"FAIL: [n_wires]={n_wires}, dynamic VS "
                    f"static run with "
                    f"[wires_per_block]={wires_per_block}. "
                    f"Params: {name}"
                )
                raise AssertionError

            q_layer.static_off()


def static_mode_dynamic_vs_static_test():
    bsz = 7

    for n_wires in range(2, 10):
        x = torch.randn((1, 100000), dtype=F_DTYPE)
        q_dev0 = tq.QuantumDevice(n_wires=n_wires)
        q_dev1 = tq.QuantumDevice(n_wires=n_wires)
        q_dev0.reset_all_eq_states(bsz)
        q_layer = QLayer(
            n_wires=n_wires,
            wires=list(range(n_wires)),
            n_ops_rd=500,
            n_ops_cin=500,
            n_funcs=500,
        )
        q_layer(q_dev0, x)
        state_dyna = q_dev0.states.clone()

        for wires_per_block in range(1, n_wires + 1):
            q_dev1.reset_all_eq_states(bsz)
            q_layer.static_off()
            q_layer.static_on(wires_per_block=wires_per_block)
            q_layer(q_dev1, x)
            state_static = q_dev1.states.clone()

            try:
                assert torch.allclose(state_dyna, state_static, atol=1e-5)
                logger.info(
                    f"PASS: [n_wires]={n_wires}, dynamic VS static "
                    f"run with [wires_per_block]={wires_per_block}"
                )
            except AssertionError:
                logger.exception(
                    f"FAIL: [n_wires]={n_wires}, dynamic VS "
                    f"static run with "
                    f"[wires_per_block]={wires_per_block}"
                )
                raise AssertionError

            q_layer.static_off()


def static_mode_dynamic_vs_static_vs_get_unitary_test():
    for n_wires in range(2, 11):
        q_dev0 = tq.QuantumDevice(n_wires=n_wires)
        q_dev1 = tq.QuantumDevice(n_wires=n_wires)
        q_dev2 = tq.QuantumDevice(n_wires=n_wires)
        q_dev0.reset_identity_states()
        q_dev1.reset_identity_states()

        q_layer = tq.RandomLayerAllTypes(n_ops=500, wires=list(range(n_wires)))
        q_layer(q_dev0)
        u_dyna = q_dev0.states.clone().reshape(2**n_wires, 2**n_wires).permute(1, 0)

        u_static_get = q_layer.get_unitary(q_dev1)
        try:
            assert torch.allclose(u_dyna, u_static_get, atol=1e-5)
            logger.info(f"PASS: [n_wires]={n_wires}, dynamic VS get_unitary()")
        except AssertionError:
            logger.exception(f"FAIL: [n_wires]={n_wires}, dynamic VS " f"get_unitary()")
            raise AssertionError

        for wires_per_block in range(1, n_wires + 1):
            q_dev2.reset_identity_states()
            q_layer.static_on(wires_per_block=wires_per_block)
            q_layer(q_dev2)
            u_static_run = (
                q_dev2.states.clone().reshape(2**n_wires, 2**n_wires).permute(1, 0)
            )
            q_layer.static_off()
            try:
                assert torch.allclose(u_dyna, u_static_run, atol=1e-5)
                logger.info(
                    f"PASS: [n_wires]={n_wires}, dynamic VS static "
                    f"run with [wires_per_block]={wires_per_block}"
                )
            except AssertionError:
                logger.exception(
                    f"FAIL: [n_wires]={n_wires}, dynamic VS "
                    f"static run with [wires_per_block]"
                    f"={wires_per_block}"
                )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--pdb", action="store_true", help="pdb")
    args = parser.parse_args()

    if args.pdb:
        pdb.set_trace()

    torch.manual_seed(42)
    np.random.seed(42)

    static_mode_dynamic_vs_static_vs_get_unitary_test()
    static_mode_dynamic_vs_static_test()
    static_mode_dynamic_vs_static_gradients_test()



import random

import torch
import torchquantum as tq
import numpy as np
import matplotlib.pyplot as plt
from .macro import C_DTYPE, ABC, ABC_ARRAY, INV_SQRT2

from typing import Union, List
from collections import Counter, OrderedDict

from torchquantum.functional import mat_dict

__all__ = [
    "expval_joint_analytical",
    "expval",
    "MeasureAll",
    "MeasureMultipleTimes",
    "MeasureMultiPauliSum",
    "MeasureMultiQubitPauliSum",
    "gen_bitstrings",
    "measure",
]


def expval_joint_analytical(
    qdev: tq.QuantumDevice,
    observable: str,
):
    """
    Compute the expectation value of a joint observable in analytical way, assuming the
    statevector is available.
    Args:
        qdev: the quantum device
        observable: the joint observable, on the qubit 0, 1, 2, 3, etc in this order
    Returns:
        the expectation value
    Examples:
    >>> import torchquantum as tq
    >>> import torchquantum.functional as tqf
    >>> x = tq.QuantumDevice(n_wires=2)
    >>> tqf.hadamard(x, wires=0)
    >>> tqf.x(x, wires=1)
    >>> tqf.cnot(x, wires=[0, 1])
    >>> print(expval_joint_analytical(x, 'II'))
    tensor([[1.0000]])
    >>> print(expval_joint_analytical(x, 'XX'))
    tensor([[1.0000]])
    >>> print(expval_joint_analytical(x, 'ZZ'))
    tensor([[-1.0000]])
    """
    # compute the hamiltonian matrix
    paulix = mat_dict["paulix"]
    pauliy = mat_dict["pauliy"]
    pauliz = mat_dict["pauliz"]
    iden = mat_dict["i"]
    pauli_dict = {"X": paulix, "Y": pauliy, "Z": pauliz, "I": iden}

    observable = observable.upper()
    assert len(observable) == qdev.n_wires
    states = qdev.get_states_1d()

    hamiltonian = pauli_dict[observable[0]].to(states.device)
    for op in observable[1:]:
        hamiltonian = torch.kron(hamiltonian, pauli_dict[op].to(states.device))

    # torch.mm(states, torch.mm(hamiltonian, states.conj().transpose(0, 1))).real

    return (
        (states.conj() * torch.mm(hamiltonian, states.transpose(0, 1)).transpose(0, 1))
        .sum(-1)
        .real
    )


def expval(
    qdev: tq.QuantumDevice,
    wires: Union[int, List[int]],
    observables: Union[tq.Observable, List[tq.Observable]],
):

    all_dims = np.arange(qdev.states.dim())
    if isinstance(wires, int):
        wires = [wires]
        observables = [observables]

    # rotation to the desired basis
    for wire, observable in zip(wires, observables):
        for rotation in observable.diagonalizing_gates():
            rotation(qdev, wires=wire)

    states = qdev.states
    # compute magnitude
    state_mag = torch.abs(states) ** 2

    expectations = []
    for wire, observable in zip(wires, observables):
        # compute marginal magnitude
        reduction_dims = np.delete(all_dims, [0, wire + 1])
        if reduction_dims.size == 0:
            probs = state_mag
        else:
            probs = state_mag.sum(list(reduction_dims))
        res = probs.mv(observable.eigvals.real.to(probs.device))
        expectations.append(res)

    return torch.stack(expectations, dim=-1)


class MeasureAll(tq.QuantumModule):
    """Obtain the expectation value of all the qubits.
    """
    def __init__(self, obs, v_c_reg_mapping=None):
        super().__init__()
        self.obs = obs
        self.v_c_reg_mapping = v_c_reg_mapping

    def forward(self, qdev: tq.QuantumDevice):
        x = expval(qdev, list(range(qdev.n_wires)), [self.obs()] * qdev.n_wires)

        if self.v_c_reg_mapping is not None:
            c2v_mapping = self.v_c_reg_mapping["c2v"]
            """
            the measurement is not normal order, need permutation
            """
            perm = []
            for k in range(x.shape[-1]):
                if k in c2v_mapping.keys():
                    perm.append(c2v_mapping[k])
            x = x[:, perm]

        if self.noise_model_tq is not None and self.noise_model_tq.is_add_noise:
            return self.noise_model_tq.apply_readout_error(x)
        else:
            return x

    def set_v_c_reg_mapping(self, mapping):
        self.v_c_reg_mapping = mapping


class MeasureMultipleTimes(tq.QuantumModule):
    """
    obs list:
    list of dict: example
    [{'wires': [0, 2, 3, 1], 'observables': ['x', 'y', 'z', 'i']
    },
    {'wires': [0, 2, 3, 1], 'observables': ['x', 'y', 'z', 'i']
    },
    ]
    """

    def __init__(self, obs_list, v_c_reg_mapping=None):
        super().__init__()
        self.obs_list = obs_list
        self.v_c_reg_mapping = v_c_reg_mapping

    def forward(self, qdev: tq.QuantumDevice):
        res_all = []

        for layer in self.obs_list:
            # create a new q device for each time of measurement
            qdev_new = tq.QuantumDevice(n_wires=qdev.n_wires)
            qdev_new.clone_states(existing_states=qdev.states)
            qdev_new.state = qdev.state

            observables = []
            for wire in range(qdev.n_wires):
                observables.append(tq.I())

            for wire, observable in zip(layer["wires"], layer["observables"]):
                observables[wire] = tq.op_name_dict[observable]()

            res = expval(
                qdev_new,
                wires=list(range(qdev.n_wires)),
                observables=observables,
            )

            if self.v_c_reg_mapping is not None:
                c2v_mapping = self.v_c_reg_mapping["c2v"]
                """
                the measurement is not normal order, need permutation
                """
                perm = []
                for k in range(res.shape[-1]):
                    if k in c2v_mapping.keys():
                        perm.append(c2v_mapping[k])
                res = res[:, perm]
            res_all.append(res)

        return torch.cat(res_all)

    def set_v_c_reg_mapping(self, mapping):
        self.v_c_reg_mapping = mapping


class MeasureMultiPauliSum(tq.QuantumModule):
    """
    similar to qiskit.opflow PauliSumOp
    obs list:
    list of dict: example
    [{'wires': [0, 2, 3, 1],
    'observables': ['x', 'y', 'z', 'i'],
    'coefficient': [1, 0.5, 0.4, 0.3]
    },
    {'wires': [0, 2, 3, 1],
    'observables': ['x', 'y', 'z', 'i'],
    'coefficient': [1, 0.5, 0.4, 0.3]
    },
    ]
    """

    def __init__(self, obs_list, v_c_reg_mapping=None):
        super().__init__()
        self.obs_list = obs_list
        self.v_c_reg_mapping = v_c_reg_mapping
        self.measure_multiple_times = MeasureMultipleTimes(
            obs_list=obs_list, v_c_reg_mapping=v_c_reg_mapping
        )

    def forward(self, qdev: tq.QuantumDevice):
        res_all = self.measure_multiple_times(qdev)

        return res_all.sum(-1)


class MeasureMultiQubitPauliSum(tq.QuantumModule):
    """obs list:
    list of dict: example
    [{'coefficient': [0.5, 0.2]},
    {'wires': [0, 2, 3, 1],
    'observables': ['x', 'y', 'z', 'i'],
    },
    {'wires': [0, 2, 3, 1],
    'observables': ['y', 'x', 'z', 'i'],
    },
    ]
    Measures 0.5 * <x y z i> + 0.2 * <y x z i>
    """

    def __init__(self, obs_list, v_c_reg_mapping=None):
        super().__init__()
        self.obs_list = obs_list
        self.v_c_reg_mapping = v_c_reg_mapping
        self.measure_multiple_times = MeasureMultipleTimes(
            obs_list=obs_list[1:], v_c_reg_mapping=v_c_reg_mapping
        )

    def forward(self, qdev: tq.QuantumDevice):
        res_all = self.measure_multiple_times(qdev)
        return (res_all * self.obs_list[0]["coefficient"]).sum(-1)


def gen_bitstrings(n_wires):
    return ["{:0{}b}".format(k, n_wires) for k in range(2**n_wires)]


def measure(qdev, n_shots=1024, draw_id=None):
    """Measure the target state and obtain classical bitstream distribution
    Args:
        q_state: input tq.QuantumDevice
        n_shots: number of simulated shots
        draw_id: which state to draw
    Returns:
        distribution of bitstrings
    """
    bitstring_candidates = gen_bitstrings(qdev.n_wires)

    state_mag = qdev.get_states_1d().abs().detach().cpu().numpy()
    distri_all = []

    for state_mag_one in state_mag:
        state_prob_one = np.abs(state_mag_one) ** 2
        measured = random.choices(
            population=bitstring_candidates,
            weights=state_prob_one,
            k=n_shots,
        )
        counter = Counter(measured)
        counter.update({key: 0 for key in bitstring_candidates})
        distri = dict(counter)
        distri = OrderedDict(sorted(distri.items()))
        distri_all.append(distri)

    if draw_id is not None:
        plt.bar(distri_all[draw_id].keys(), distri_all[draw_id].values())
        plt.xticks(rotation="vertical")
        plt.xlabel("bitstring [qubit0, qubit1, ..., qubitN]")
        plt.title("distribution of measured bitstrings")
        plt.show()
    return distri_all



import copy
import sched
import qiskit
import itertools
import numpy as np

from itertools import repeat
from qiskit.providers import aer
from qiskit.providers.fake_provider import *
from qiskit.circuit import Gate
from qiskit.compiler import assemble
from qiskit import pulse, QuantumCircuit, IBMQ
from qiskit.pulse.instructions import Instruction
from qiskit.pulse.transforms import block_to_schedule
from qiskit_nature.drivers import UnitsType, Molecule
from scipy.optimize import minimize, LinearConstraint
from qiskit_nature.converters.second_quantization import QubitConverter
from qiskit_nature.properties.second_quantization.electronic import ParticleNumber
from qiskit_nature.problems.second_quantization import ElectronicStructureProblem
from typing import List, Tuple, Iterable, Union, Dict, Callable, Set, Optional, Any
from qiskit.pulse import (
    Schedule,
    GaussianSquare,
    Drag,
    Delay,
    Play,
    ControlChannel,
    DriveChannel,
)
from qiskit_nature.mappers.second_quantization import ParityMapper, JordanWignerMapper
from qiskit_nature.transformers.second_quantization.electronic import (
    ActiveSpaceTransformer,
)
from qiskit_nature.drivers.second_quantization import (
    ElectronicStructureDriverType,
    ElectronicStructureMoleculeDriver,
)


def is_parametric_pulse(t0, *inst: Union["Schedule", Instruction]):
    inst = t0[1]
    t0 = t0[0]
    if isinstance(inst, pulse.Play):
        return True
    else:
        return False


def extract_ampreal(pulse_prog):
    # extract the real part of pulse amplitude, igonred the imaginary part.
    amp_list = list(
        map(
            lambda x: x[1].pulse.amp,
            pulse_prog.filter(is_parametric_pulse).instructions,
        )
    )
    amp_list = np.array(amp_list)
    ampa_list = np.angle(np.array(amp_list))
    return ampa_list


def extract_amp(pulse_prog):
    # extract the pulse amplitdue.
    amp_list = list(
        map(
            lambda x: x[1].pulse.amp,
            pulse_prog.filter(is_parametric_pulse).instructions,
        )
    )
    amp_list = np.array(amp_list)
    ampa_list = np.angle(np.array(amp_list))
    ampn_list = np.abs(np.array(amp_list))
    amps_list = []
    for i, j in zip(ampn_list, ampa_list):
        amps_list.append(i)
        amps_list.append(j)
    amps_list = np.array(amps_list)
    return amps_list


def is_phase_pulse(t0, *inst: Union["Schedule", Instruction]):
    inst = t0[1]
    t0 = t0[0]
    if isinstance(inst, pulse.ShiftPhase):
        return True
    return False


def extract_phase(pulse_prog):

    for _, ShiftPhase in pulse_prog.filter(is_phase_pulse).instructions:
        # print(play.pulse.amp)
        pass
    instructions = pulse_prog.filter(is_phase_pulse).instructions

    phase_list = list(
        map(
            lambda x: x[1]._operands[0],
            pulse_prog.operands[0].filter(is_phase_pulse).instructions,
        )
    )
    return phase_list


def cir2pul(circuit, backend):
    # transform quantum circuit to pulse schedule
    with pulse.build(backend) as pulse_prog:
        pulse.call(circuit)
    return pulse_prog


def snp(qubit, backend):
    circuit = QuantumCircuit(qubit + 1)
    circuit.h(qubit)
    sched = cir2pul(circuit, backend)
    sched = block_to_schedule(sched)
    return sched


def tnp(qubit, cqubit, backend):
    circuit = QuantumCircuit(cqubit + 1)
    circuit.cx(qubit, cqubit)
    sched = cir2pul(circuit, backend)
    sched = block_to_schedule(sched)
    return sched


def pul_append(sched1, sched2):
    sched = sched1.append(sched2)
    return sched


def map_amp(pulse_ansatz, modified_list):
    sched = Schedule()
    for inst, amp in zip(
        pulse_ansatz.filter(is_parametric_pulse).instructions, modified_list
    ):
        inst[1].pulse._amp = amp
    for i in pulse_ansatz.instructions:
        if is_parametric_pulse(i):
            sched += copy.deepcopy(i[1])
    return sched


def get_from(d: dict, key: str):

    value = 0
    if key in d:
        value = d[key]
    return value


def run_pulse_sim(measurement_pulse):
    measure_result = []
    for measure_pulse in measurement_pulse:
        shots = 1024
        pulse_sim = qiskit.providers.aer.PulseSimulator.from_backend(FakeJakarta())
        pul_sim = assemble(
            measure_pulse,
            backend=pulse_sim,
            shots=1024,
            meas_level=2,
            meas_return="single",
        )
        results = pulse_sim.run(pul_sim).result()

        counts = results.get_counts()
        expectation_value = (
            (get_from(counts, "00") + get_from(counts, "11"))
            - (get_from(counts, "10") + get_from(counts, "01"))
        ) / shots
        measure_result.append(expectation_value)
    return measure_result


def gen_LC(parameters_array):
    dim_design = int(len(parameters_array))
    Mid = int(len(parameters_array) / 2)
    bound = np.ones((dim_design, 2)) * np.array([0, 0.9])
    bound[-Mid:] = bound[-Mid:] * np.pi * 2
    tol = 1e-3  # tolerance for optimization precision.
    lb = bound[:, 0]
    ub = bound[:, 1]
    LC = LinearConstraint(np.eye(dim_design), lb, ub, keep_feasible=False)
    return LC


def observe_genearte(pulse_ansatz, backend):
    qubits = 0, 1
    with pulse.build(backend) as pulse_measurez0:
        # z measurement of qubit 0 and 1
        pulse.call(pulse_ansatz)
        pulse.barrier(0, 1)
        pulse.measure(0)
    with pulse.build(backend) as pulse_measurez1:
        # z measurement of qubit 0 and 1
        pulse.call(pulse_ansatz)
        pulse.barrier(0, 1)
        pulse.measure(1)
    with pulse.build(backend) as pulse_measurez:
        # z measurement of qubit 0 and 1
        pulse.call(pulse_ansatz)
        pulse.barrier(0, 1)
        pulse.measure(qubits)
    with pulse.build(backend) as pulse_measurex:
        # x measurement of qubit 0 and 1
        pulse.call(pulse_ansatz)
        pulse.barrier(0, 1)
        pulse.u2(0, np.pi, 0)
        pulse.u2(0, np.pi, 1)
        pulse.barrier(0, 1)
        pulse.measure(qubits)

    with pulse.build(backend) as pulse_measurey:
        # y measurement of qubit 0 and 1
        pulse.call(pulse_ansatz)
        pulse.barrier(0, 1)
        pulse.u2(np.pi / 2, 0, 0)
        pulse.u2(np.pi / 2, 0, 1)
        pulse.barrier(0, 1)
        pulse.measure(qubits)
    measurement_pulse = [
        pulse_measurez0,
        pulse_measurez1,
        pulse_measurez,
        pulse_measurex,
    ]
    return measurement_pulse



# FIXED parameterization

# This code is part of Qiskit.
#
# (C) Copyright IBM 2018, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Qiskit Aer qasm simulator backend.
"""

import copy
import datetime
import logging
import time
import uuid
import warnings
from abc import ABC, abstractmethod

from qiskit.circuit import QuantumCircuit, ParameterExpression, Delay
from qiskit.compiler import assemble
from qiskit.providers import BackendV1 as Backend
from qiskit.providers.models import BackendStatus
from qiskit.pulse import Schedule, ScheduleBlock
from qiskit.qobj import QasmQobj, PulseQobj
from qiskit.result import Result
from qiskit.utils import deprecate_arguments
from ..aererror import AerError
from ..jobs import AerJob, AerJobSet, split_qobj
from ..noise.noise_model import NoiseModel, QuantumErrorLocation
from ..noise.errors.quantum_error import QuantumChannelInstruction
from .aer_compiler import compile_circuit
from .backend_utils import format_save_type, circuit_optypes

# Logger
logger = logging.getLogger(__name__)


class AerBackend(Backend, ABC):
    """Qiskit Aer Backend class."""

    def __init__(
        self,
        configuration,
        properties=None,
        defaults=None,
        backend_options=None,
        provider=None,
    ):
        """Aer class for backends.

        This method should initialize the module and its configuration, and
        raise an exception if a component of the module is
        not available.

        Args:
            configuration (BackendConfiguration): backend configuration.
            properties (BackendProperties or None): Optional, backend properties.
            defaults (PulseDefaults or None): Optional, backend pulse defaults.
            provider (Provider): Optional, provider responsible for this backend.
            backend_options (dict or None): Optional set custom backend options.

        Raises:
            AerError: if there is no name in the configuration
        """
        # Init configuration and provider in Backend
        configuration.simulator = True
        configuration.local = True
        super().__init__(configuration, provider=provider)

        # Initialize backend properties and pulse defaults.
        self._properties = properties
        self._defaults = defaults

        # Custom option values for config, properties, and defaults
        self._options_configuration = {}
        self._options_defaults = {}
        self._options_properties = {}

        # Set options from backend_options dictionary
        if backend_options is not None:
            self.set_options(**backend_options)

    def _convert_circuit_binds(self, circuit, binds):
        parameterizations = []
        for index, inst_tuple in enumerate(circuit.data):
            if inst_tuple[0].is_parameterized():
                for bind_pos, param in enumerate(inst_tuple[0].params):
                    if param in binds:
                        parameterizations.append([[index, bind_pos], binds[param]])
                    elif isinstance(param, ParameterExpression):
                        # If parameter expression has no unbound parameters
                        # it's already bound and should be skipped
                        if not param.parameters:
                            continue
                        local_binds = {
                            k: v for k, v in binds.items() if k in param.parameters
                        }
                        bind_list = [
                            dict(zip(local_binds, t))
                            for t in zip(*local_binds.values())
                        ]
                        bound_values = [float(param.bind(x)) for x in bind_list]
                        parameterizations.append([[index, bind_pos], bound_values])
        return parameterizations

    def _convert_binds(self, circuits, parameter_binds):
        if isinstance(circuits, QuantumCircuit):
            if len(parameter_binds) > 1:
                raise AerError(
                    "More than 1 parameter table provided for a single circuit"
                )

            return [self._convert_circuit_binds(circuits, parameter_binds[0])]
        elif len(parameter_binds) != len(circuits):
            raise AerError(
                "Number of input circuits does not match number of input "
                "parameter bind dictionaries"
            )
        parameterizations = [
            self._convert_circuit_binds(circuit, parameter_binds[idx])
            for idx, circuit in enumerate(circuits)
        ]
        return parameterizations

    # pylint: disable=arguments-differ
    @deprecate_arguments({"qobj": "circuits"})
    def run(self, circuits, validate=False, parameter_binds=None, **run_options):
        """Run a qobj on the backend.

        Args:
            circuits (QuantumCircuit or list): The QuantumCircuit (or list
                of QuantumCircuit objects) to run
            validate (bool): validate the Qobj before running (default: False).
            parameter_binds (list): A list of parameter binding dictionaries.
                                    See additional information (default: None).
            run_options (kwargs): additional run time backend options.

        Returns:
            AerJob: The simulation job.

        Raises:
            AerError: If ``parameter_binds`` is specified with a qobj input or has a
                length mismatch with the number of circuits.

        Additional Information:
            * Each parameter binding dictionary is of the form::

                {
                    param_a: [val_1, val_2],
                    param_b: [val_3, val_1],
                }

              for all parameters in that circuit. The length of the value
              list must be the same for all parameters, and the number of
              parameter dictionaries in the list must match the length of
              ``circuits`` (if ``circuits`` is a single ``QuantumCircuit``
              object it should a list of length 1).
            * kwarg options specified in ``run_options`` will temporarily override
              any set options of the same name for the current run.

        Raises:
            ValueError: if run is not implemented
        """
        if isinstance(circuits, (QasmQobj, PulseQobj)):
            warnings.warn(
                "Using a qobj for run() is deprecated as of qiskit-aer 0.9.0"
                " and will be removed no sooner than 3 months from that release"
                " date. Transpiled circuits should now be passed directly using"
                " `backend.run(circuits, **run_options).",
                DeprecationWarning,
                stacklevel=2,
            )
            if parameter_binds:
                raise AerError("Parameter binds can't be used with an input qobj")
            # A work around to support both qobj options and run options until
            # qobj is deprecated is to copy all the set qobj.config fields into
            # run_options that don't override existing fields. This means set
            # run_options fields will take precidence over the value for those
            # fields that are set via assemble.
            if not run_options:
                run_options = circuits.config.__dict__
            else:
                run_options = copy.copy(run_options)
                for key, value in circuits.config.__dict__.items():
                    if key not in run_options and value is not None:
                        run_options[key] = value
            qobj = self._assemble(circuits, **run_options)
        else:
            qobj = self._assemble(
                circuits, parameter_binds=parameter_binds, **run_options
            )

        # Optional validation
        if validate:
            self._validate(qobj)

        # Get executor from qobj config and delete attribute so qobj can still be serialized
        executor = getattr(qobj.config, "executor", None)
        if hasattr(qobj.config, "executor"):
            delattr(qobj.config, "executor")

        # Optionally split the job
        experiments = split_qobj(
            qobj,
            max_size=getattr(qobj.config, "max_job_size", None),
            max_shot_size=getattr(qobj.config, "max_shot_size", None),
        )

        # Temporarily remove any executor from options so that job submission
        # can work with Dask client executors which can't be pickled
        opts_executor = getattr(self._options, "executor", None)
        if hasattr(self._options, "executor"):
            self._options.executor = None

        # Submit job
        job_id = str(uuid.uuid4())
        if isinstance(experiments, list):
            aer_job = AerJobSet(self, job_id, self._run, experiments, executor)
        else:
            aer_job = AerJob(self, job_id, self._run, experiments, executor)
        aer_job.submit()

        # Restore removed executor after submission
        if hasattr(self._options, "executor"):
            self._options.executor = opts_executor

        return aer_job

    def configuration(self):
        """Return the simulator backend configuration.

        Returns:
            BackendConfiguration: the configuration for the backend.
        """
        config = copy.copy(self._configuration)
        for key, val in self._options_configuration.items():
            setattr(config, key, val)
        # If config has custom instructions add them to
        # basis gates to include them for the terra transpiler
        if hasattr(config, "custom_instructions"):
            config.basis_gates = config.basis_gates + config.custom_instructions
        return config

    def properties(self):
        """Return the simulator backend properties if set.

        Returns:
            BackendProperties: The backend properties or ``None`` if the
                               backend does not have properties set.
        """
        properties = copy.copy(self._properties)
        for key, val in self._options_properties.items():
            setattr(properties, key, val)
        return properties

    def defaults(self):
        """Return the simulator backend pulse defaults.

        Returns:
            PulseDefaults: The backend pulse defaults or ``None`` if the
                           backend does not support pulse.
        """
        defaults = copy.copy(self._defaults)
        for key, val in self._options_defaults.items():
            setattr(defaults, key, val)
        return defaults

    @classmethod
    def _default_options(cls):
        pass

    def clear_options(self):
        """Reset the simulator options to default values."""
        self._options = self._default_options()
        self._options_configuration = {}
        self._options_properties = {}
        self._options_defaults = {}

    def status(self):
        """Return backend status.

        Returns:
            BackendStatus: the status of the backend.
        """
        return BackendStatus(
            backend_name=self.name(),
            backend_version=self.configuration().backend_version,
            operational=True,
            pending_jobs=0,
            status_msg="",
        )

    def _run(self, qobj, job_id="", format_result=True):
        """Run a job"""
        # Start timer
        start = time.time()

        # Run simulation
        output = self._execute(qobj)

        # Validate output
        if not isinstance(output, dict):
            logger.error("%s: simulation failed.", self.name())
            if output:
                logger.error("Output: %s", output)
            raise AerError("simulation terminated without returning valid output.")

        # Format results
        output["job_id"] = job_id
        output["date"] = datetime.datetime.now().isoformat()
        output["backend_name"] = self.name()
        output["backend_version"] = self.configuration().backend_version

        # Add execution time
        output["time_taken"] = time.time() - start

        # Display warning if simulation failed
        if not output.get("success", False):
            msg = "Simulation failed"
            if "status" in output:
                msg += f" and returned the following error message:\n{output['status']}"
            logger.warning(msg)
        if format_result:
            return self._format_results(output)
        return output

    @staticmethod
    def _format_results(output):
        """Format C++ simulator output for constructing Result"""
        for result in output["results"]:
            data = result.get("data", {})
            metadata = result.get("metadata", {})
            save_types = metadata.get("result_types", {})
            save_subtypes = metadata.get("result_subtypes", {})
            for key, val in data.items():
                if key in save_types:
                    data[key] = format_save_type(
                        val, save_types[key], save_subtypes[key]
                    )
        return Result.from_dict(output)

    def _assemble(self, circuits, parameter_binds=None, **run_options):
        """Assemble one or more Qobj for running on the simulator"""
        if isinstance(circuits, (QasmQobj, PulseQobj)):
            qobj = circuits
        else:
            # Generate optypes for circuit
            # Generate opsets of instructions
            if isinstance(circuits, (QuantumCircuit, Schedule, ScheduleBlock)):
                circuits = [circuits]
            optypes = [circuit_optypes(circ) for circ in circuits]

            # Compile Qasm3 instructions
            circuits, optypes = compile_circuit(
                circuits, basis_gates=self.configuration().basis_gates, optypes=optypes
            )

            # run option noise model
            circuits, optypes, run_options = self._assemble_noise_model(
                circuits, optypes, **run_options
            )

            if parameter_binds:
                # Handle parameter binding
                # parameterizations = self._convert_binds(circuits, parameter_binds)
                # assemble_binds = []
                # assemble_binds.append({param: 1 for bind in parameter_binds for param in bind})

                # qobj = assemble(
                #     circuits,
                #     backend=self,
                #     parameter_binds=assemble_binds,
                #     parameterizations=parameterizations)
                qobj = assemble(circuits, self, parameter_binds=parameter_binds)
            else:
                qobj = assemble(circuits, backend=self)

            # Add optypes to qobj
            # We convert to strings to avoid pybinding of types
            qobj.config.optypes = [
                set(i.__name__ for i in optype) if optype else set()
                for optype in optypes
            ]

        # Add options
        for key, val in self.options.__dict__.items():
            if val is not None:
                setattr(qobj.config, key, val)

        # Override with run-time options
        for key, val in run_options.items():
            setattr(qobj.config, key, val)

        return qobj

    def _assemble_noise_model(self, circuits, optypes, **run_options):
        """Move quantum error instructions from circuits to noise model"""
        if isinstance(circuits, (QuantumCircuit, Schedule, ScheduleBlock)):
            run_circuits = [circuits]
        else:
            # Make a shallow copy so we can modify list elements if required
            run_circuits = copy.copy(circuits)

        # Flag for if we need to make a deep copy of the noise model
        # This avoids unnecessarily copying the noise model for circuits
        # that do not contain a quantum error
        updated_noise = False
        noise_model = run_options.get(
            "noise_model", getattr(self.options, "noise_model", None)
        )

        # Add custom pass noise only to QuantumCircuit objects that contain delay
        # instructions since this is the only instruction handled by the noise pass
        # at present
        if noise_model and all(
            isinstance(circ, QuantumCircuit) for circ in run_circuits
        ):
            npm = noise_model._pass_manager()
            if npm is not None:
                # Get indicies of circuits that need noise transpiling
                transpile_idxs = [
                    idx for idx, optype in enumerate(optypes) if Delay in optype
                ]

                # Transpile only the required circuits
                transpiled_circuits = npm.run([run_circuits[i] for i in transpile_idxs])
                if isinstance(transpiled_circuits, QuantumCircuit):
                    transpiled_circuits = [transpiled_circuits]

                # Update the circuits with transpiled ones
                for idx, circ in zip(transpile_idxs, transpiled_circuits):
                    run_circuits[idx] = circ
                    optypes[idx] = circuit_optypes(circ)

        # Check if circuits contain quantum error instructions
        for idx, circ in enumerate(run_circuits):
            if QuantumChannelInstruction in optypes[idx] and not isinstance(
                circ, (Schedule, ScheduleBlock)
            ):
                updated_circ = False
                new_data = []
                for inst, qargs, cargs in circ.data:
                    if isinstance(inst, QuantumChannelInstruction):
                        updated_circ = True
                        if not updated_noise:
                            # Deep copy noise model on first update
                            if noise_model is None:
                                noise_model = NoiseModel()
                            else:
                                noise_model = copy.deepcopy(noise_model)
                            updated_noise = True
                        # Extract error and replace with place holder
                        qerror = inst._quantum_error
                        qerror_loc = QuantumErrorLocation(qerror)
                        new_data.append((qerror_loc, qargs, cargs))
                        optypes[idx].add(QuantumErrorLocation)
                        # Add error to noise model
                        if qerror.id not in noise_model._default_quantum_errors:
                            noise_model.add_all_qubit_quantum_error(qerror, qerror.id)
                    else:
                        new_data.append((inst, qargs, cargs))
                if updated_circ:
                    new_circ = circ.copy()
                    new_circ.data = new_data
                    run_circuits[idx] = new_circ
                    optypes[idx].discard(QuantumChannelInstruction)

        # If we modified the existing noise model, add it to the run options
        if updated_noise:
            run_options["noise_model"] = noise_model

        # Return the possibly updated circuits and noise model
        return run_circuits, optypes, run_options

    def _get_executor(self, **run_options):
        """Get the executor"""
        if "executor" in run_options:
            return run_options["executor"]
        else:
            return getattr(self._options, "executor", None)

    @abstractmethod
    def _execute(self, qobj):
        """Execute a qobj on the backend.

        Args:
            qobj (QasmQobj or PulseQobj): simulator input.

        Returns:
            dict: return a dictionary of results.
        """
        pass

    def _validate(self, qobj):
        """Validate the qobj for the backend"""
        pass

    def set_option(self, key, value):
        """Special handling for setting backend options.

        This method should be extended by sub classes to
        update special option values.

        Args:
            key (str): key to update
            value (any): value to update.

        Raises:
            AerError: if key is 'method' and val isn't in available methods.
        """
        # Add all other options to the options dict
        # TODO: in the future this could be replaced with an options class
        #       for the simulators like configuration/properties to show all
        #       available options
        if hasattr(self._configuration, key):
            self._set_configuration_option(key, value)
        elif hasattr(self._properties, key):
            self._set_properties_option(key, value)
        elif hasattr(self._defaults, key):
            self._set_defaults_option(key, value)
        else:
            if not hasattr(self._options, key):
                raise AerError("Invalid option %s" % key)
            if value is not None:
                # Only add an option if its value is not None
                setattr(self._options, key, value)
            else:
                # If setting an existing option to None reset it to default
                # this is for backwards compatibility when setting it to None would
                # remove it from the options dict
                setattr(self._options, key, getattr(self._default_options(), key))

    def set_options(self, **fields):
        """Set the simulator options"""
        for key, value in fields.items():
            self.set_option(key, value)

    def _set_configuration_option(self, key, value):
        """Special handling for setting backend configuration options."""
        if value is not None:
            self._options_configuration[key] = value
        elif key in self._options_configuration:
            self._options_configuration.pop(key)

    def _set_properties_option(self, key, value):
        """Special handling for setting backend properties options."""
        if value is not None:
            self._options_properties[key] = value
        elif key in self._options_properties:
            self._options_properties.pop(key)

    def _set_defaults_option(self, key, value):
        """Special handling for setting backend defaults options."""
        if value is not None:
            self._options_defaults[key] = value
        elif key in self._options_defaults:
            self._options_defaults.pop(key)

    def __repr__(self):
        """String representation of an AerBackend."""
        name = self.__class__.__name__
        display = f"'{self.name()}'"
        return f"{name}({display})"



import torch
import torchquantum as tq
from qiskit import pulse, QuantumCircuit
from qiskit.pulse import library
from qiskit.test.mock import FakeQuito, FakeArmonk, FakeBogota
from qiskit.compiler import assemble, schedule
from .qiskit_macros import IBMQ_PNAMES


def circ2pulse(circuits, name):
    if name in IBMQ_PNAMES:
        backend = name()
        with pulse.build(backend) as pulse_tq:
            qc = circuits
            qc.measure_all()
            pulse.call(qc)
        pulse_tq.draw()



import torch

from torchpack.datasets.dataset import Dataset
from torchvision import datasets, transforms
from typing import List
from torchpack.utils.logging import logger
from torchvision.transforms import InterpolationMode


__all__ = ["MNIST"]


resize_modes = {
    "bilinear": InterpolationMode.BILINEAR,
    "bicubic": InterpolationMode.BICUBIC,
    "nearest": InterpolationMode.NEAREST,
}


class MNISTDataset:
    def __init__(
        self,
        root: str,
        split: str,
        train_valid_split_ratio: List[float],
        center_crop,
        resize,
        resize_mode,
        binarize,
        binarize_threshold,
        digits_of_interest,
        n_test_samples,
        n_valid_samples,
        fashion,
        n_train_samples,
        same_n_samples_each_class=False,
    ):
        self.root = root
        self.split = split
        self.train_valid_split_ratio = train_valid_split_ratio
        self.data = None
        self.center_crop = center_crop
        self.resize = resize
        self.resize_mode = resize_modes[resize_mode]
        self.binarize = binarize
        self.binarize_threshold = binarize_threshold
        self.digits_of_interest = digits_of_interest
        self.n_test_samples = n_test_samples
        self.n_valid_samples = n_valid_samples
        self.n_train_samples = n_train_samples
        self.fashion = fashion

        self.n_digits = len(digits_of_interest)

        self.same_n_samples_each_class = same_n_samples_each_class

        self.load()
        self.n_instance = len(self.data)

    def _get_indices(self, subset, n_samples):
        sample_ctr = {}
        sample_quota = {}
        indices = []
        n_samples_each_class = n_samples // self.n_digits

        for digit_of_interest in self.digits_of_interest:
            sample_ctr[digit_of_interest] = 0
            if digit_of_interest == self.digits_of_interest[-1]:
                sample_quota[digit_of_interest] = (
                    n_samples - (self.n_digits - 1) * n_samples_each_class
                )
            else:
                sample_quota[digit_of_interest] = n_samples_each_class

        for idx in subset.indices:
            digit = subset.dataset[idx][1]
            if sample_ctr[digit] < sample_quota[digit]:
                indices.append(idx)
                sample_ctr[digit] += 1

        return indices

    def load(self):
        tran = [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
        if not self.center_crop == 28:
            tran.append(transforms.CenterCrop(self.center_crop))
        if not self.resize == 28:
            tran.append(transforms.Resize(self.resize, interpolation=self.resize_mode))
        transform = transforms.Compose(tran)

        if self.split == "train" or self.split == "valid":
            if self.fashion:
                train_valid = datasets.FashionMNIST(
                    self.root, train=True, download=True, transform=transform
                )
            else:
                train_valid = datasets.MNIST(
                    self.root, train=True, download=True, transform=transform
                )
            idx, _ = torch.stack(
                [train_valid.targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            train_valid.targets = train_valid.targets[idx]
            train_valid.data = train_valid.data[idx]

            train_len = int(self.train_valid_split_ratio[0] * len(train_valid))
            split = [train_len, len(train_valid) - train_len]
            train_subset, valid_subset = torch.utils.data.random_split(
                train_valid, split, generator=torch.Generator().manual_seed(1)
            )

            if self.split == "train":
                if self.n_train_samples is None:
                    # use all samples in train set
                    self.data = train_subset
                else:
                    if self.same_n_samples_each_class:
                        train_subset.indices = self._get_indices(
                            train_subset, self.n_train_samples
                        )
                    else:
                        train_subset.indices = train_subset.indices[
                            : self.n_train_samples
                        ]
                    self.data = train_subset
                    logger.warning(
                        f"Only use the front "
                        f"{self.n_train_samples} images as "
                        f"TRAIN set."
                    )
            else:
                if self.n_valid_samples is None:
                    # use all samples in valid set
                    self.data = valid_subset
                else:
                    # use a subset of valid set, useful to speedup evo search
                    if self.same_n_samples_each_class:
                        valid_subset.indices = self._get_indices(
                            valid_subset, self.n_valid_samples
                        )
                    else:
                        valid_subset.indices = valid_subset.indices[
                            : self.n_valid_samples
                        ]
                    self.data = valid_subset
                    logger.warning(
                        f"Only use the front "
                        f"{self.n_valid_samples} images as "
                        f"VALID set."
                    )

        else:
            if self.fashion:
                test = datasets.FashionMNIST(
                    self.root, train=False, transform=transform
                )
            else:
                test = datasets.MNIST(self.root, train=False, transform=transform)
            idx, _ = torch.stack(
                [test.targets == number for number in self.digits_of_interest]
            ).max(dim=0)
            test.targets = test.targets[idx]
            test.data = test.data[idx]
            if self.n_test_samples is None:
                # use all samples as test set
                self.data = test
            else:
                # use a subset as test set
                if self.same_n_samples_each_class:
                    sample_ctr = {}
                    sample_quota = {}
                    indices = []
                    n_samples_each_class = self.n_test_samples // self.n_digits

                    for digit_of_interest in self.digits_of_interest:
                        sample_ctr[digit_of_interest] = 0
                        if digit_of_interest == self.digits_of_interest[-1]:
                            sample_quota[digit_of_interest] = (
                                self.n_test_samples
                                - (self.n_digits - 1) * n_samples_each_class
                            )
                        else:
                            sample_quota[digit_of_interest] = n_samples_each_class

                    for idx, target in enumerate(test.targets):
                        digit = target.item()
                        if sample_ctr[digit] < sample_quota[digit]:
                            indices.append(idx)
                            sample_ctr[digit] += 1

                    test.targets = test.targets[indices]
                    test.data = test.data[indices]
                else:
                    test.targets = test.targets[: self.n_test_samples]
                    test.data = test.data[: self.n_test_samples]

                self.data = test
                logger.warning(
                    f"Only use the front {self.n_test_samples} " f"images as TEST set."
                )

    def __getitem__(self, index: int):
        img = self.data[index][0]
        if self.binarize:
            img = 1.0 * (img > self.binarize_threshold) + -1.0 * (
                img <= self.binarize_threshold
            )

        digit = self.digits_of_interest.index(self.data[index][1])
        instance = {"image": img, "digit": digit}
        return instance

    def __len__(self) -> int:
        return self.n_instance


class MNIST(Dataset):
    def __init__(
        self,
        root: str,
        train_valid_split_ratio: List[float],
        center_crop=28,
        resize=28,
        resize_mode="bilinear",
        binarize=False,
        binarize_threshold=0.1307,
        digits_of_interest=tuple(range(10)),
        n_test_samples=None,
        n_valid_samples=None,
        fashion=False,
        n_train_samples=None,
    ):
        self.root = root

        super().__init__(
            {
                split: MNISTDataset(
                    root=root,
                    split=split,
                    train_valid_split_ratio=train_valid_split_ratio,
                    center_crop=center_crop,
                    resize=resize,
                    resize_mode=resize_mode,
                    binarize=binarize,
                    binarize_threshold=binarize_threshold,
                    digits_of_interest=digits_of_interest,
                    n_test_samples=n_test_samples,
                    n_valid_samples=n_valid_samples,
                    fashion=fashion,
                    n_train_samples=n_train_samples,
                )
                for split in ["train", "valid", "test"]
            }
        )


if __name__ == "__main__":
    import pdb

    pdb.set_trace()
    mnist = MNISTDataset(
        root="../fashion_mnist_data",
        split="train",
        train_valid_split_ratio=[0.9, 0.1],
        center_crop=28,
        resize=28,
        resize_mode="bilinear",
        binarize=False,
        binarize_threshold=0.1307,
        digits_of_interest=(3, 6),
        n_test_samples=100,
        n_valid_samples=1000,
        fashion=True,
        n_train_samples=10000,
    )
    mnist.__getitem__(20)
    print("finish")



import torch
import torch.nn as nn
import numpy as np

from typing import Union, List, Iterable


__all__ = [
    "QuantumPulse",
    "QuantumPulseGaussian",
    "QuantumPulseDirect",
]


class QuantumPulse(nn.Module):
    """
    The Quantum Pulse simulator
    """

    def __init__(self):
        super().__init__()
        pass

    def forward(self):
        pass


class QuantumPulseDirect(QuantumPulse):
    def __init__(
        self,
        n_steps: int,
        hamil,
        delta_t: float = 1.0,
        initial_shape: List[float] = None,
    ):
        super().__init__()
        self.hamil = torch.tensor(hamil, dtype=torch.complex64)
        if initial_shape is not None:
            assert len(initial_shape) == n_steps
            initial_shape = torch.Tensor(initial_shape)
        else:
            initial_shape = torch.ones(n_steps)
        self.pulse_shape = nn.Parameter(initial_shape)
        self.n_steps = n_steps
        self.delta_t = delta_t

    def get_unitary(self):
        unitary_per_step = []
        for k in range(self.n_steps):
            magnitude = self.pulse_shape[k]
            unitary = torch.matrix_exp(-1j * self.hamil * magnitude * self.delta_t)
            # print(unitary @ unitary.conj().T)
            # unitary_mag = (unitary[0]**2).sum().sqrt()
            # unitary = unitary_mag / unitary

            unitary_per_step.append(unitary)

        u_overall = None
        for k, u in enumerate(unitary_per_step):
            if not k:
                u_overall = u
            else:
                u_overall = u_overall @ u

        return u_overall

    def __repr__(self):
        return f"QuantumPulse Direct \n shape: {self.pulse_shape}"


class QuantumPulseGaussian(QuantumPulse):
    """Gaussian Quantum Pulse, will only count +- five sigmas"""

    def __init__(
        self,
        hamil,
        n_steps: int = 100,
        delta_t: float = 1.0,
        x_min: float = -10,
        x_max: float = 10,
        initial_params: List[float] = None,
    ):
        super(QuantumPulseGaussian, self).__init__()
        self.hamil = torch.tensor(hamil, dtype=torch.complex64)
        self.delta_t = delta_t
        # mag, mu, sigma
        if initial_params is not None:
            assert len(initial_params) == 3
            initial_params = torch.Tensor(initial_params)
        else:
            initial_params = torch.ones(3)

        self.pulse_params = nn.Parameter(initial_params)
        self.n_steps = n_steps
        self.delta_x = (x_max - x_min) / n_steps
        self.x_list = torch.tensor(np.arange(x_min, x_max, self.delta_x))

    def get_unitary(self):
        self.mag = self.pulse_params[0]
        self.mu = self.pulse_params[1]
        self.sigma = self.pulse_params[2]

        # delta_x = (10 * self.sigma / self.n_steps).item()
        # self.x_list = torch.tensor(np.arange(
        # (self.mu - 5 * self.sigma).item(),
        # (self.mu + 5 * self.sigma).item(), delta_x))

        self.pulse_shape = self.mag * torch.exp(
            -((self.x_list - self.mu) ** 2) / (2 * self.sigma**2)
        )

        unitary_per_step = []
        for k in range(self.n_steps):
            magnitude = self.pulse_shape[k]
            unitary = torch.matrix_exp(
                -1j * self.hamil * magnitude * self.delta_t * self.delta_x
            )
            # print(unitary @ unitary.conj().T)
            # unitary_mag = (unitary[0]**2).sum().sqrt()
            # unitary = unitary_mag / unitary

            unitary_per_step.append(unitary)

        u_overall = None
        for k, u in enumerate(unitary_per_step):
            if not k:
                u_overall = u
            else:
                u_overall = u_overall @ u

        return u_overall

    def __repr__(self):
        return f"QuantumPulse Guassian \n shape: {self.pulse_shape}"


if __name__ == "__main__":
    import pdb

    pdb.set_trace()
    pulse = QuantumPulseDirect(n_steps=10, hamil=[[0, 1], [1, 0]])

    print(pulse.get_unitary())

    print("finish")



import torch.nn as nn
import torchquantum as tq
import torchquantum.functional as tqf
import numpy as np


from typing import Iterable
from torchquantum.plugins.qiskit_macros import QISKIT_INCOMPATIBLE_FUNC_NAMES
from torchpack.utils.logging import logger

__all__ = [
    "QuantumModuleFromOps",
    "TrainableOpAll",
    "ClassicalInOpAll",
    "FixedOpAll",
    "TwoQAll",
    "RandomLayer",
    "RandomLayerAllTypes",
    "Op1QAllLayer",
    "Op2QAllLayer",
    "Op2QButterflyLayer",
    "Op2QDenseLayer",
    "layer_name_dict",
    "CXLayer",
    "CXCXCXLayer",
    "SWAPSWAPLayer",
    "RXYZCXLayer0",
]


class QuantumModuleFromOps(tq.QuantumModule):
    def __init__(self, ops):
        super().__init__()
        self.ops = tq.QuantumModuleList(ops)

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        self.q_device = q_device
        for op in self.ops:
            op(q_device)


class TrainableOpAll(tq.QuantumModule):
    """Rotation rx on all qubits
    The rotation angle is a parameter of each rotation gate
    One potential optimization is to compute the unitary of all gates
    together.
    """

    def __init__(self, n_gate: int, op: tq.Operation):
        super().__init__()
        self.n_gate = n_gate
        self.gate_all = nn.ModuleList()
        for k in range(self.n_gate):
            self.gate_all.append(op(has_params=True, trainable=True))

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        # rx on all wires, assert the number of gate is the same as the number
        # of wires in the device.
        assert self.n_gate == q_device.n_wires, (
            f"Number of rx gates ({self.n_gate}) is different from number "
            f"of wires ({q_device.n_wires})!"
        )

        for k in range(self.n_gate):
            self.gate_all[k](q_device, wires=k)


class ClassicalInOpAll(tq.QuantumModule):
    def __init__(self, n_gate: int, op: tq.Operator):
        super().__init__()
        self.n_gate = n_gate
        self.gate_all = nn.ModuleList()
        for k in range(self.n_gate):
            self.gate_all.append(op())

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice, x):
        # rx on all wires, assert the number of gate is the same as the number
        # of wires in the device.
        assert self.n_gate == q_device.n_wires, (
            f"Number of rx gates ({self.n_gate}) is different from number "
            f"of wires ({q_device.n_wires})!"
        )

        for k in range(self.n_gate):
            self.gate_all[k](q_device, wires=k, params=x[:, k])


class FixedOpAll(tq.QuantumModule):
    def __init__(self, n_gate: int, op: tq.Operator):
        super().__init__()
        self.n_gate = n_gate
        self.gate_all = nn.ModuleList()
        for k in range(self.n_gate):
            self.gate_all.append(op())

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        # rx on all wires, assert the number of gate is the same as the number
        # of wires in the device.
        assert self.n_gate == q_device.n_wires, (
            f"Number of rx gates ({self.n_gate}) is different from number "
            f"of wires ({q_device.n_wires})!"
        )

        for k in range(self.n_gate):
            self.gate_all[k](q_device, wires=k)


class TwoQAll(tq.QuantumModule):
    def __init__(self, n_gate: int, op: tq.Operator):
        super().__init__()
        self.n_gate = n_gate
        self.op = op()

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        for k in range(self.n_gate - 1):
            self.op(q_device, wires=[k, k + 1])
        self.op(q_device, wires=[self.n_gate - 1, 0])


class RandomLayer(tq.QuantumModule):
    def __init__(
        self,
        wires,
        n_ops=None,
        n_params=None,
        op_ratios=None,
        op_types=(tq.RX, tq.RY, tq.RZ, tq.CNOT),
        seed=None,
        qiskit_compatible=False,
    ):
        super().__init__()
        self.n_ops = n_ops
        self.n_params = n_params
        assert n_params is not None or n_ops is not None
        self.wires = wires if isinstance(wires, Iterable) else [wires]
        self.n_wires = len(wires)

        op_types = op_types if isinstance(op_types, Iterable) else [op_types]
        if op_ratios is None:
            op_ratios = [1] * len(op_types)
        else:
            op_ratios = op_ratios if isinstance(op_ratios, Iterable) else [op_ratios]
        op_types_valid = []
        op_ratios_valid = []

        if qiskit_compatible:
            for op_type, op_ratio in zip(op_types, op_ratios):
                if op_type().name.lower() in QISKIT_INCOMPATIBLE_FUNC_NAMES:
                    logger.warning(
                        f"Remove {op_type} from op_types to make "
                        f"the layer qiskit-compatible."
                    )
                else:
                    op_types_valid.append(op_type)
                    op_ratios_valid.append(op_ratio)
        else:
            op_types_valid = op_types
            op_ratios_valid = op_ratios

        self.op_types = op_types_valid
        self.op_ratios = np.array(op_ratios_valid) / sum(op_ratios_valid)

        self.seed = seed
        self.op_list = tq.QuantumModuleList()
        if seed is not None:
            np.random.seed(seed)
        self.build_random_layer()

    def rebuild_random_layer_from_op_list(self, n_ops_in, wires_in, op_list_in):
        """Used for loading random layer from checkpoint"""
        self.n_ops = n_ops_in
        self.wires = wires_in
        self.op_list = tq.QuantumModuleList()
        for op_in in op_list_in:
            op = tq.op_name_dict[op_in.name.lower()](
                has_params=op_in.has_params,
                trainable=op_in.trainable,
                wires=op_in.wires,
                n_wires=op_in.n_wires,
            )
            self.op_list.append(op)

    def build_random_layer(self):
        op_cnt = 0
        param_cnt = 0
        while True:
            op = np.random.choice(self.op_types, p=self.op_ratios)
            n_op_wires = op.num_wires
            if n_op_wires > self.n_wires:
                continue
            if n_op_wires == -1:
                is_AnyWire = True
                n_op_wires = self.n_wires
            else:
                is_AnyWire = False

            op_wires = list(
                np.random.choice(self.wires, size=n_op_wires, replace=False)
            )
            if is_AnyWire:
                if op().name in ["MultiRZ"]:
                    operation = op(
                        has_params=True,
                        trainable=True,
                        n_wires=n_op_wires,
                        wires=op_wires,
                    )
                else:
                    operation = op(n_wires=n_op_wires, wires=op_wires)
            elif op().name in tq.Operator.parameterized_ops:
                operation = op(has_params=True, trainable=True, wires=op_wires)
            else:
                operation = op(wires=op_wires)
            self.op_list.append(operation)
            op_cnt += 1
            param_cnt += op.num_params

            if self.n_ops is not None and op_cnt == self.n_ops:
                break
            elif self.n_ops is None and self.n_params is not None:
                if param_cnt == self.n_params:
                    break
                elif param_cnt > self.n_params:
                    """
                    the last operation has too many params and exceed the
                    constraint, so need to remove it and sample another
                    """
                    op_cnt -= 1
                    param_cnt -= op.num_params
                    del self.op_list[-1]

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        for op in self.op_list:
            op(q_device)


class RandomLayerAllTypes(RandomLayer):
    def __init__(
        self,
        wires,
        n_ops=None,
        n_params=None,
        op_ratios=None,
        op_types=(
            tq.Hadamard,
            tq.SHadamard,
            tq.PauliX,
            tq.PauliY,
            tq.PauliZ,
            tq.S,
            tq.T,
            tq.SX,
            tq.CNOT,
            tq.CZ,
            tq.CY,
            tq.RX,
            tq.RY,
            tq.RZ,
            tq.RZZ,
            tq.SWAP,
            tq.CSWAP,
            tq.Toffoli,
            tq.PhaseShift,
            tq.Rot,
            tq.MultiRZ,
            tq.CRX,
            tq.CRY,
            tq.CRZ,
            tq.CRot,
            tq.U1,
            tq.U2,
            tq.U3,
            tq.MultiCNOT,
            tq.MultiXCNOT,
        ),
        seed=None,
        qiskit_compatible=False,
    ):
        super().__init__(
            wires=wires,
            n_ops=n_ops,
            n_params=n_params,
            op_ratios=op_ratios,
            op_types=op_types,
            seed=seed,
            qiskit_compatible=qiskit_compatible,
        )


class SimpleQLayer(tq.QuantumModule):
    def __init__(self, n_wires):
        super().__init__()
        self.n_wires = n_wires
        self.gate1 = tq.RX(has_params=True, trainable=True)
        self.gate2 = tq.RY(has_params=True, trainable=True)
        self.gate3 = tq.RZ(has_params=True, trainable=True)

    @tq.static_support
    def forward(self, q_dev):
        self.q_device = q_dev
        tqf.x(q_dev, wires=0, static=self.static_mode, parent_graph=self.graph)
        self.gate1(q_dev, wires=1)
        self.gate2(q_dev, wires=1)
        self.gate3(q_dev, wires=1)
        tqf.x(q_dev, wires=2, static=self.static_mode, parent_graph=self.graph)


class CXLayer(tq.QuantumModule):
    def __init__(self, n_wires):
        super().__init__()
        self.n_wires = n_wires

    @tq.static_support
    def forward(self, q_dev):
        self.q_device = q_dev
        tqf.cnot(q_dev, wires=[0, 1], static=self.static_mode, parent_graph=self.graph)


class CXCXCXLayer(tq.QuantumModule):
    def __init__(self, n_wires):
        super().__init__()
        self.n_wires = n_wires

    @tq.static_support
    def forward(self, q_dev):
        self.q_device = q_dev
        tqf.cnot(q_dev, wires=[0, 1], static=self.static_mode, parent_graph=self.graph)
        tqf.cnot(q_dev, wires=[1, 2], static=self.static_mode, parent_graph=self.graph)
        tqf.cnot(q_dev, wires=[2, 0], static=self.static_mode, parent_graph=self.graph)


class SWAPSWAPLayer(tq.QuantumModule):
    def __init__(self, n_wires):
        super().__init__()
        self.n_wires = n_wires

    @tq.static_support
    def forward(self, q_dev):
        self.q_device = q_dev
        tqf.swap(q_dev, wires=[0, 1], static=self.static_mode, parent_graph=self.graph)
        tqf.swap(q_dev, wires=[1, 2], static=self.static_mode, parent_graph=self.graph)


class Op1QAllLayer(tq.QuantumModule):
    def __init__(self, op, n_wires: int, has_params=False, trainable=False):
        super().__init__()
        self.n_wires = n_wires
        self.op = op
        self.ops_all = tq.QuantumModuleList()
        for k in range(n_wires):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(self.n_wires):
            self.ops_all[k](q_device, wires=k)


class Op2QAllLayer(tq.QuantumModule):
    """pattern:
    circular = False
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5]
    jump = 3: [0, 3], [1, 4], [2, 5]
    jump = 4: [0, 4], [1, 5]
    jump = 5: [0, 5]

    circular = True
    jump = 1: [0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0]
    jump = 2: [0, 2], [1, 3], [2, 4], [3, 5], [4, 0], [5, 1]
    jump = 3: [0, 3], [1, 4], [2, 5], [3, 0], [4, 1], [5, 2]
    jump = 4: [0, 4], [1, 5], [2, 0], [3, 1], [4, 2], [5, 3]
    jump = 5: [0, 5], [1, 0], [2, 1], [3, 2], [4, 3], [5, 4]
    """

    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
        circular=False,
    ):
        super().__init__()
        self.n_wires = n_wires
        self.jump = jump
        self.circular = circular
        self.op = op
        self.ops_all = tq.QuantumModuleList()

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        if circular:
            n_ops = n_wires
        else:
            n_ops = n_wires - jump
        for k in range(n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(len(self.ops_all)):
            wires = [k, (k + self.jump) % self.n_wires]
            if self.wire_reverse:
                wires.reverse()
            self.ops_all[k](q_device, wires=wires)


class Op2QFit32Layer(tq.QuantumModule):
    def __init__(
        self,
        op,
        n_wires: int,
        has_params=False,
        trainable=False,
        wire_reverse=False,
        jump=1,
        circular=False,
    ):
        super().__init__()
        self.n_wires = n_wires
        self.jump = jump
        self.circular = circular
        self.op = op
        self.ops_all = tq.QuantumModuleList()

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        # if circular:
        #     n_ops = n_wires
        # else:
        #     n_ops = n_wires - jump
        n_ops = 32
        for k in range(n_ops):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    @tq.static_support
    def forward(self, q_device):
        for k in range(len(self.ops_all)):
            wires = [k % self.n_wires, (k + self.jump) % self.n_wires]
            if self.wire_reverse:
                wires.reverse()
            self.ops_all[k](q_device, wires=wires)


class Op2QButterflyLayer(tq.QuantumModule):
    """pattern: [0, 5], [1, 4], [2, 3]"""

    def __init__(
        self, op, n_wires: int, has_params=False, trainable=False, wire_reverse=False
    ):
        super().__init__()
        self.n_wires = n_wires
        self.op = op
        self.ops_all = tq.QuantumModuleList()

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        for k in range(n_wires // 2):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    def forward(self, q_device):
        for k in range(len(self.ops_all)):
            wires = [k, self.n_wires - 1 - k]
            if self.wire_reverse:
                wires.reverse()
            self.ops_all[k](q_device, wires=wires)


class Op2QDenseLayer(tq.QuantumModule):
    """pattern:
    [0, 1], [0, 2], [0, 3], [0, 4], [0, 5]
    [1, 2], [1, 3], [1, 4], [1, 5]
    [2, 3], [2, 4], [2, 5]
    [3, 4], [3, 5]
    [4, 5]
    """

    def __init__(
        self, op, n_wires: int, has_params=False, trainable=False, wire_reverse=False
    ):
        super().__init__()
        self.n_wires = n_wires
        self.op = op
        self.ops_all = tq.QuantumModuleList()

        # reverse the wires, for example from [1, 2] to [2, 1]
        self.wire_reverse = wire_reverse

        for k in range(self.n_wires * (self.n_wires - 1) // 2):
            self.ops_all.append(op(has_params=has_params, trainable=trainable))

    def forward(self, q_device):
        k = 0
        for i in range(self.n_wires - 1):
            for j in range(i + 1, self.n_wires):
                wires = [i, j]
                if self.wire_reverse:
                    wires.reverse()
                self.ops_all[k](q_device, wires=wires)
                k += 1


class LayerTemplate0(tq.QuantumModule):
    def __init__(self, arch: dict = None):
        super().__init__()
        self.n_wires = arch["n_wires"]
        self.arch = arch

        self.n_blocks = arch.get("n_blocks", None)
        self.n_layers_per_block = arch.get("n_layers_per_block", None)

        self.layers_all = self.build_layers()

    def build_layers(self):
        raise NotImplementedError

    @tq.static_support
    def forward(self, q_device: tq.QuantumDevice):
        self.q_device = q_device
        for k in range(len(self.layers_all)):
            self.layers_all[k](q_device)


class U3CU3Layer0(LayerTemplate0):
    """u3 cu3 blocks"""

    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.U3, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op2QAllLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return layers_all


class CU3Layer0(LayerTemplate0):
    """u3 cu3 blocks"""

    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QAllLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=False,
                )
            )
        return layers_all


class CXRZSXLayer0(LayerTemplate0):
    """CXRZSX blocks"""

    def build_layers(self):
        layers_all = tq.QuantumModuleList()

        layers_all.append(
            Op1QAllLayer(
                op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
            )
        )
        layers_all.append(
            Op2QAllLayer(op=tq.CNOT, n_wires=self.n_wires, jump=1, circular=False)
        )
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.SX, n_wires=self.n_wires, has_params=False, trainable=False
                )
            )
        layers_all.append(
            Op1QAllLayer(
                op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
            )
        )
        return layers_all


class SethLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
        return layers_all


class SethLayer1(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return layers_all


class SethLayer2(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QFit32Layer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return layers_all


class RZZLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RZZ,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return layers_all


class BarrenLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        layers_all.append(
            Op1QAllLayer(
                op=tq.SHadamard,
                n_wires=self.n_wires,
            )
        )
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RX, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(Op2QAllLayer(op=tq.CZ, n_wires=self.n_wires, jump=1))
        return layers_all


class FarhiLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RZX,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
            layers_all.append(
                Op2QAllLayer(
                    op=tq.RXX,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )
        return layers_all


class MaxwellLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RX, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(Op1QAllLayer(op=tq.S, n_wires=self.n_wires))
            layers_all.append(
                Op2QAllLayer(op=tq.CNOT, n_wires=self.n_wires, jump=1, circular=True)
            )

            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(Op1QAllLayer(op=tq.T, n_wires=self.n_wires))
            layers_all.append(
                Op2QAllLayer(op=tq.SWAP, n_wires=self.n_wires, jump=1, circular=True)
            )

            layers_all.append(
                Op1QAllLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(Op1QAllLayer(op=tq.Hadamard, n_wires=self.n_wires))
            layers_all.append(
                Op2QAllLayer(op=tq.SSWAP, n_wires=self.n_wires, jump=1, circular=True)
            )

            layers_all.append(
                Op1QAllLayer(
                    op=tq.U1, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op2QAllLayer(
                    op=tq.CU3,
                    n_wires=self.n_wires,
                    has_params=True,
                    trainable=True,
                    jump=1,
                    circular=True,
                )
            )

        return layers_all


class RYRYCXLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(CXLayer(n_wires=self.n_wires))
        return layers_all


class RYRYRYCXCXCXLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(CXCXCXLayer(n_wires=self.n_wires))
        return layers_all


class RYRYRYLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
        return layers_all


class RYRYRYSWAPSWAPLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(SWAPSWAPLayer(n_wires=self.n_wires))
        return layers_all


class SWAPSWAPLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(SWAPSWAPLayer(n_wires=self.n_wires))
        return layers_all


class RXYZCXLayer0(LayerTemplate0):
    def build_layers(self):
        layers_all = tq.QuantumModuleList()
        for k in range(self.arch["n_blocks"]):
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RX, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RY, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op1QAllLayer(
                    op=tq.RZ, n_wires=self.n_wires, has_params=True, trainable=True
                )
            )
            layers_all.append(
                Op2QAllLayer(op=tq.CNOT, n_wires=self.n_wires, jump=1, circular=True)
            )
        return layers_all


layer_name_dict = {
    "u3cu3_0": U3CU3Layer0,
    "cu3_0": CU3Layer0,
    "cxrzsx_0": CXRZSXLayer0,
    "seth_0": SethLayer0,
    "seth_1": SethLayer1,
    "seth_2": SethLayer2,
    "rzz_0": RZZLayer0,
    "barren_0": BarrenLayer0,
    "farhi_0": FarhiLayer0,
    "maxwell_0": MaxwellLayer0,
    "ryrycx": RYRYCXLayer0,
    "ryryrycxcxcx": RYRYRYCXCXCXLayer0,
    "ryryry": RYRYRYLayer0,
    "swapswap": SWAPSWAPLayer0,
    "ryryryswapswap": RYRYRYSWAPSWAPLayer0,
    "rxyzcx_0": RXYZCXLayer0,
}



import torchquantum as tq

QISKIT_INCOMPATIBLE_OPS = [
    tq.Rot,
    tq.MultiRZ,
    tq.CRot,
]

QISKIT_INCOMPATIBLE_FUNC_NAMES = [
    "rot",
    "multirz",
    "crot",
]

IBMQ_NAMES = [
    "ibmqx2",
    "ibmq_16_melbourne",
    "ibmq_5_yorktown",
    "ibmq_armonk",
    "ibmq_athens",
    "ibmq_belem",
    "ibmq_bogota",
    "ibmq_casablanca",
    "ibmq_dublin",
    "ibmq_guadalupe",
    "ibmq_jakarta",
    "ibmq_kolkata",
    "ibmq_lima",
    "ibmq_manhattan",
    "ibmq_manila",
    "ibmq_montreal",
    "ibmq_mumbai",
    "ibmq_paris",
    "ibmq_qasm_simulator",
    "ibmq_quito",
    "ibmq_rome",
    "ibmq_santiago",
    "ibmq_sydney",
    "ibmq_toronto",
    "simulator_extended_stabilizer",
    "simulator_mps",
    "simulator_stabilizer",
    "simulator_statevector",
    "ibm_auckland",
    "ibm_cairo",
    "ibm_geneva",
    "ibm_hanoi",
    "ibm_ithaca",
    "ibm_lagos",
    "ibm_nairobi",
    "ibm_oslo",
    "ibm_peekskill",
    "ibm_perth",
    "ibm_washington",
]

IBMQ_PNAMES = [
    "FakeArmonk",
    "FakeBogota" "FakeQuito",
]



import numpy as np
import time
from scipy.optimize import minimize
from sklearn.gaussian_process import GaussianProcessRegressor as GPR
from scipy.optimize import LinearConstraint
from sklearn.gaussian_process.kernels import RBF, ConstantKernel, Matern, WhiteKernel
from scipy.stats import norm


def acquisition(x_scaled, hyper_param, model, min_Y):  # x_scaled: 1 * dim
    x_scaled = x_scaled.reshape(1, -1)
    if "LCB" in hyper_param[0]:
        mean, std = model.predict(x_scaled, return_std=True)
        return mean[0] - hyper_param[1] * std[0]
    elif hyper_param[0] == "EI":
        tau = min_Y
        mean, std = model.predict(x_scaled, return_std=True)
        tau_scaled = (tau - mean) / std
        res = (tau - mean) * norm.cdf(tau_scaled) + std * norm.pdf(tau_scaled)
        return -res  # maximize Ei = minimize -EI
    elif hyper_param[0] == "PI":
        tau = min_Y
        mean, std = model.predict(x_scaled, return_std=True)
        tau_scaled = (tau - mean) / std
        res = norm.cdf(tau_scaled)
        return -res
    else:
        raise ValueError("acquisition function is not implemented")


def bayes_opt(
    func,
    dim_design,
    N_sim,
    N_initial,
    w_bound,
    hyper_param,
    store=False,
    verbose=True,
    file_suffix="",
):
    """

    :param func: [functional handle], represents the objective function. objective = func(design)
    :param dim_design: [int], the dimension of the design variable
    :param N_sim: [int], The total number of allowable simulations
    :param N_initial: [int], The number of simulations used to set up the initial dataset
    :param w_bound: [(dim_design, 2) np.array], the i-th row contains the lower bound and upper bound for the i-th variable
    :param hyper_param: the parameter for the acquisition function e.g., ['LCB','0.3'], ['EI'], ['PI']
    :param verbose: [Bool], if it is true, print detailed information in each iteration of Bayesian optimization
    :param file_suffix: [string], file suffix used in storing optimization information
    :return:
    cur_best_w: [(dim_design,) np.array], the best design variable
    cur_best_y: [float], the minimum objective value
    """

    # initialization: set up the training dataset X, Y.
    print("Begin initializing...")
    X = np.repeat(
        (w_bound[:, 1] - w_bound[:, 0]).reshape(1, -1), N_initial, axis=0
    ) * np.random.rand(N_initial, dim_design) + np.repeat(
        w_bound[:, 0].reshape(1, -1), N_initial, axis=0
    )
    Y = np.zeros((N_initial,))

    for i in range(N_initial):
        Y[i] = func(X[i, :])
        print(
            "Simulate the %d-th sample... with metric: %.3e" % (i, Y[i])
        ) if verbose else None
    print("Finish initialization with best metric: %.3e" % (np.min(Y)))

    # define several working variables, will be used to store results
    pred_mean = np.zeros(N_sim - N_initial)
    pred_std = np.zeros(N_sim - N_initial)
    acq_list = np.zeros(N_sim - N_initial)

    # Goes into real Bayesian Optimization
    cur_count, cur_best_w, cur_best_y = N_initial, None, 1e10
    while cur_count < N_sim:

        # build gaussian process on the normalized data
        wrk_mean, wrk_std = X.mean(axis=0), X.std(axis=0)
        model = GPR(
            kernel=ConstantKernel(1, (1e-9, 1e9)) * RBF(1.0, (1e-5, 1e5)),
            normalize_y=True,
            n_restarts_optimizer=100,
        )
        model.fit(np.divide(X - wrk_mean, wrk_std), Y)

        # define acquisition function, np.min(Y) is needed in EI and PI, but not LCB
        acq_func = lambda x_scaled: acquisition(x_scaled, hyper_param, model, np.min(Y))

        # optimize the acquisition function independently for N_inner times, select the best one
        N_inner, cur_min, opt = 20, np.inf, None
        for i in range(N_inner):
            w_init = (w_bound[:, 1] - w_bound[:, 0]) * np.random.rand(dim_design) + (
                w_bound[:, 0]
            )
            LC = LinearConstraint(
                np.eye(dim_design),
                np.divide(w_bound[:, 0] - wrk_mean, wrk_std),
                np.divide(w_bound[:, 1] - wrk_mean, wrk_std),
                keep_feasible=False,
            )
            cur_opt = minimize(
                acq_func,
                np.divide(w_init - wrk_mean, wrk_std),
                method="COBYLA",
                constraints=LC,
                options={"disp": False},
            )
            wrk = acq_func(cur_opt.x)
            if cur_min >= wrk:
                cur_min = wrk
                opt = cur_opt

        # do a clipping to avoid violation of constraints (just in case), and also undo the normalization
        newX = np.clip(opt.x * wrk_std + wrk_mean, w_bound[:, 0], w_bound[:, 1])
        star_time = time.time()
        cur_count += 1
        newY = func(newX)
        end_time = time.time()
        X, Y = np.concatenate((X, newX.reshape(1, -1)), axis=0), np.concatenate(
            (Y, [newY]), axis=0
        )

        # save and display information
        ind = np.argmin(Y)
        cur_predmean, cur_predstd = model.predict(
            (np.divide(newX - wrk_mean, wrk_std)).reshape(1, -1), return_std=True
        )
        cur_acq = acq_func(np.divide(newX - wrk_mean, wrk_std))
        cur_best_w, cur_best_y = X[ind, :], Y[ind]
        pred_mean[cur_count - N_initial - 1], pred_std[cur_count - N_initial - 1] = (
            cur_predmean,
            cur_predstd,
        )
        acq_list[cur_count - N_initial - 1] = cur_acq
        if store:
            np.save("./result/X_" + file_suffix + ".npy", X)
            np.save("./result/Y_" + file_suffix + ".npy", Y)
            np.save("./result/cur_best_w_" + file_suffix + ".npy", cur_best_w)
            np.save("./result/cur_best_y_" + file_suffix + ".npy", cur_best_y)
            np.save("./result/pred_mean_" + file_suffix + ".npy", pred_mean)
            np.save("./result/pred_std_" + file_suffix + ".npy", pred_std)
            np.save("./result/acq_list_" + file_suffix + ".npy", acq_list)
        if verbose:
            print("-" * 10)
            print("Number of function evaluations: %d" % cur_count)
            print("Optimize acq message: ", opt.message)
            print(
                "Model predict(new sampled X)... mean: %.3e, std:%.3e"
                % (cur_predmean, cur_predstd)
            )
            print("Acq(new sampled X): %.3e" % cur_acq)
            print(
                "Y(new sampled X): %.3e, simulation time: %.3e"
                % (newY, end_time - star_time)
            )
            print("Current best design: ", cur_best_w)
            print("Current best function value: %.3e" % cur_best_y)

    return cur_best_w, cur_best_y


if __name__ == "__main__":
    seed = 0
    np.random.seed(seed)

    # example: minimize x1^2 + x2^2 + x3^2 + ...
    dim_design = 10
    N_total = 200
    N_initial = 40
    bound = np.ones((dim_design, 2)) * np.array([-10, 10])  # -inf < xi < inf

    func = lambda x: np.sum(x * x)
    cur_best_w, cur_best_y = bayes_opt(
        func,
        dim_design,
        N_total,
        N_initial,
        bound,
        ["LCB", 0.3],
        store=False,
        verbose=True,
        file_suffix=str(seed),
    )

    print(cur_best_w)
    print(cur_best_y)



# from setuptools import setup, find_packages
#
# VERSION = {}  # type: ignore
#
# with open("torchquantum/__version__.py", "r") as version_file:
#     exec(version_file.read(), VERSION)
#
# if __name__ == "__main__":
#     setup(
#         name="torchquantum",
#         version=VERSION["version"],
#         description="A PyTorch-based framework for differentiable classical simulation of quantum computing",
#         url="https://github.com/mit-han-lab/torchquantum",
#         author="Hanrui Wang, Jiannan Cao, Jessica Ding, Jiai Gu, Song Han, Zirui Li, Zhiding Liang, Pengyu Liu, Mohammadreza Tavasoli",
#         author_email="hanruiwang.hw@gmail.com",
#         license="MIT",
#         install_requires=[
#             "numpy>=1.19.2",
#             "torchvision>=0.9.0.dev20210130",
#             "tqdm>=4.56.0",
#             "setuptools>=52.0.0",
#             "torch>=1.8.0",
#             "torchpack>=0.3.0",
#             "qiskit>=0.32.1",
#             "matplotlib>=3.3.2",
#             "pathos>=0.2.7",
#             "pylatexenc>=2.10",
#             "dill==0.3.4",
#         ],
#         extras_require={"doc": ["nbsphinx", "recommonmark"]},
#         python_requires=">=3.5",
#         include_package_data=True,
#         packages=find_packages(),
#     )

version = "0.1.7"



from torchpack.datasets.dataset import Dataset
import numpy as np
import torch

__all__ = ["LayerRegression"]


class LayerRegressionDatasetV:
    def __init__(self):
        self.data = None
        self.n_instance = 10000
        self.input = np.random.rand(5).astype(np.float) * 2 * np.pi
        self.output = np.random.rand(5).astype(np.float) * 2 - 1

    def __getitem__(self, index: int):
        instance = {"input": self.input, "output": self.output}
        return instance

    def __len__(self) -> int:
        return self.n_instance


class LayerRegressionV(Dataset):
    def __init__(
        self,
    ):
        super().__init__(
            {split: LayerRegressionDatasetV() for split in ["train", "valid", "test"]}
        )


class LayerRegressionDataset:
    def __init__(self):
        self.data = None
        self.n_instance = 10000
        mat = torch.randn((2**5, 2**5), dtype=torch.complex64)
        mat = mat.svd()[0].data

        self.output = mat

    def __getitem__(self, index: int):
        instance = {"input": self.output, "output": self.output}
        return instance

    def __len__(self) -> int:
        return self.n_instance


class LayerRegression(Dataset):
    def __init__(
        self,
    ):
        super().__init__(
            {split: LayerRegressionDataset() for split in ["train", "valid", "test"]}
        )



import torch
import torch.nn.functional as F


def complex_mse(output, target):
    return F.mse_loss(torch.view_as_real(output), torch.view_as_real(target))


def complex_mae(output, target):
    return (torch.view_as_real(output) - torch.view_as_real(target)).abs().mean()


def minimize(output, target):
    return output.sum()


def maximize(output, target):
    return -output.sum()
